from abc import ABC, abstractmethod
from typing import Dict, List, Callable

import tensorflow as tf

from ml4h.TensorMap import TensorMap

Tensor = tf.Tensor
Block = Callable[[Tensor, Dict[TensorMap, List[Tensor]]], Tensor]


class Block(ABC):
    @abstractmethod
    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        pass

    def can_apply(self):
        return True



from typing import Dict

import cv2
import h5py
import logging
import numpy as np
from scipy import ndimage

from ml4h.TensorMap import TensorMap
from ml4h.normalizer import ZeroMeanStd1
from ml4h.tensormap.general import get_tensor_at_last_date, pad_or_crop_array_to_shape


def find_border_components(labeled_image):
    border_components = set()

    # Check top and bottom rows
    border_components.update(np.unique(labeled_image[0, :]))
    border_components.update(np.unique(labeled_image[-1, :]))

    # Check left and right columns
    border_components.update(np.unique(labeled_image[:, 0]))
    border_components.update(np.unique(labeled_image[:, -1]))

    # Remove background (label 0)
    border_components.discard(0)

    return border_components


def erase_border_components(image, erase_value=252):
    binary_image = (image == erase_value).astype(int)

    # Find connected components and label them
    labeled_image, num_features = ndimage.label(binary_image)

    # Find connected components adjacent to the image border
    border_components = find_border_components(labeled_image)

    for label in border_components:
        image[labeled_image == label] = 0

    return image


def dxa_background_erase(tm, hd5, dependents={}):
    tensor = get_tensor_at_last_date(hd5, tm.path_prefix, tm.name)
    bc = np.bincount(tensor.flatten().astype(np.int64))
    for val, count in zip(reversed(bc.argsort()), reversed(sorted(bc))):
        if val == 252 and count > 20000:
            tensor = erase_border_components(tensor, erase_value=252)
            break
    tensor = pad_or_crop_array_to_shape(tm.shape, tensor)
    return tensor


dxa_2 = TensorMap(
    'dxa_1_2',
    shape=(768, 768, 1),
    path_prefix='ukb_dxa',
    tensor_from_file=dxa_background_erase,
    normalization=ZeroMeanStd1(),
)
dxa_5 = TensorMap(
    'dxa_1_5',
    shape=(768, 768, 1),
    path_prefix='ukb_dxa',
    tensor_from_file=dxa_background_erase,
    normalization=ZeroMeanStd1(),
)
dxa_6 = TensorMap(
    'dxa_1_6',
    shape=(864, 768, 1),
    path_prefix='ukb_dxa',
    tensor_from_file=dxa_background_erase,
    normalization=ZeroMeanStd1(),
)
dxa_7 = TensorMap(
    'dxa_1_7',
    shape=(1792, 896, 1),
    path_prefix='ukb_dxa',
    tensor_from_file=dxa_background_erase,
    normalization=ZeroMeanStd1(),
)
dxa_8 = TensorMap(
    'dxa_1_8',
    shape=(640, 768, 1),
    path_prefix='ukb_dxa',
    tensor_from_file=dxa_background_erase,
    normalization=ZeroMeanStd1(),
)
dxa_12 = TensorMap(
    'dxa_1_12',
    shape=(928, 352, 1),
    path_prefix='ukb_dxa',
    tensor_from_file=dxa_background_erase,
    normalization=ZeroMeanStd1(),
)
dxa_11 = TensorMap(
    'dxa_1_11',
    shape=(896, 352, 1),
    path_prefix='ukb_dxa',
    tensor_from_file=dxa_background_erase,
    normalization=ZeroMeanStd1(),
)


def register_to_sample(
    register_hd5, register_path, register_name, register_shape,
    number_of_iterations = 5000, termination_eps = 1e-4, warp_mode = cv2.MOTION_TRANSLATION,
):
    register_tensor = None
    # Define termination criteria
    criteria = (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, number_of_iterations, termination_eps)
    with h5py.File(register_hd5, 'r') as r_hd5:
        r_tensor = get_tensor_at_last_date(r_hd5, register_path, register_name)
        register_tensor = pad_or_crop_array_to_shape(register_shape, r_tensor)

    def registered_tensor(tm, hd5, dependents={}):
        tensor = get_tensor_at_last_date(hd5, tm.path_prefix, tm.name)
        tensor = pad_or_crop_array_to_shape(tm.shape, tensor)

        bc = np.bincount(tensor.flatten().astype(np.int64))
        for val, count in zip(reversed(bc.argsort()), reversed(sorted(bc))):
            if count > 10000:
                tensor[tensor == float(val)] = 0
            else:
                break

        if warp_mode == cv2.MOTION_HOMOGRAPHY:
            warp_matrix = np.eye(3, 3, dtype=np.float32)
        else:
            warp_matrix = np.eye(2, 3, dtype=np.float32)

        try:
            (cc, warp_matrix) = cv2.findTransformECC(
                register_tensor.astype(np.float32), tensor.astype(np.float32),
                warp_matrix, warp_mode, criteria,
            )
            if warp_mode == cv2.MOTION_HOMOGRAPHY:
                # Use warpPerspective for Homography
                tensor[..., 0] = cv2.warpPerspective(
                    tensor, warp_matrix, (register_shape[1], register_shape[0]),
                    flags=cv2.INTER_LINEAR + cv2.WARP_INVERSE_MAP,
                )
            else:
                # Use warpAffine for Translation, Euclidean and Affine
                tensor[..., 0] = cv2.warpAffine(
                    tensor, warp_matrix, (register_shape[1], register_shape[0]),
                    flags=cv2.INTER_LINEAR + cv2.WARP_INVERSE_MAP,
                );
        except cv2.error as e:
            logging.debug(f'Got cv2 error {e}')
        return tensor
    return registered_tensor


dxa_6_translate = TensorMap(
    'dxa_1_6',
    shape=(864, 736, 1),
    path_prefix='ukb_dxa',
    tensor_from_file=register_to_sample(
        register_hd5='/mnt/disks/dxa-tensors-50k/2023_06_12/1000107.hd5',
        register_path='ukb_dxa',
        register_name='dxa_1_6',
        register_shape=(864, 736, 1),
    ),
    normalization=ZeroMeanStd1(),
)

dxa_12_translate = TensorMap(
    'dxa_1_12',
    shape=(896, 320, 1),
    path_prefix='ukb_dxa',
    tensor_from_file=register_to_sample(
        register_hd5='/mnt/disks/dxa-tensors-50k/2023_06_12/1105369.hd5',
        register_path='ukb_dxa',
        register_name='dxa_1_12',
        register_shape=(896, 320, 1),
    ),
    normalization=ZeroMeanStd1(),
)


dxa_2_homography = TensorMap(
    'dxa_1_2',
    shape=(768, 768, 1),
    path_prefix='ukb_dxa',
    tensor_from_file=register_to_sample(
        register_hd5='/mnt/disks/dxa-tensors-50k/2023_06_12/1105369.hd5',
        register_path='ukb_dxa',
        register_name='dxa_1_2',
        register_shape=(768, 768, 1),
        number_of_iterations=5000,
        termination_eps=2e-4,
        warp_mode=cv2.MOTION_HOMOGRAPHY,
    ),
    normalization=ZeroMeanStd1(),
)


dxa_5_homography = TensorMap(
    'dxa_1_5',
    shape=(768, 768, 1),
    path_prefix='ukb_dxa',
    tensor_from_file=register_to_sample(
        register_hd5='/mnt/disks/dxa-tensors-50k/2023_06_12/1105369.hd5',
        register_path='ukb_dxa',
        register_name='dxa_1_5',
        register_shape=(768, 768, 1),
        number_of_iterations=5000,
        termination_eps=2e-4,
        warp_mode=cv2.MOTION_HOMOGRAPHY,
    ),
    normalization=ZeroMeanStd1(),
)

dxa_8_homography = TensorMap(
    'dxa_1_8',
    shape=(640, 768, 1),
    path_prefix='ukb_dxa',
    tensor_from_file=register_to_sample(
        register_hd5='/mnt/disks/dxa-tensors-50k/2023_06_12/1105369.hd5',
        register_path='ukb_dxa',
        register_name='dxa_1_8',
        register_shape=(640, 768, 1),
        number_of_iterations=5000,
        termination_eps=2e-4,
        warp_mode=cv2.MOTION_HOMOGRAPHY,
    ),
    normalization=ZeroMeanStd1(),
)

dxa_12_homography = TensorMap(
    'dxa_1_12',
    shape=(928, 352, 1),
    path_prefix='ukb_dxa',
    tensor_from_file=register_to_sample(
        register_hd5='/mnt/disks/dxa-tensors-50k/2023_06_12/1105369.hd5',
        register_path='ukb_dxa',
        register_name='dxa_1_12',
        register_shape=(928, 352, 1),
        number_of_iterations=2000,
        termination_eps=2e-4,
        warp_mode=cv2.MOTION_HOMOGRAPHY,
    ),
    normalization=ZeroMeanStd1(),
)


def image_from_hd5(tm: TensorMap, hd5: h5py.File, dependents: Dict = {}) -> np.ndarray:
    return np.array(hd5[tm.name][:tm.shape[0], :tm.shape[1], :tm.shape[2]], dtype=np.float32)


dxa_1_12_registered = TensorMap('dxa_1_12_registered', shape=(928, 352, 1), tensor_from_file=image_from_hd5)
dxa_1_12_flow = TensorMap('dxa_1_12_flow', shape=(928, 352, 2), tensor_from_file=image_from_hd5)



"""Methods for integration of static plots within notebooks."""
import os
import tempfile
from typing import List, Optional, Union

from IPython.display import HTML
from IPython.display import SVG
import numpy as np
from ml4h.plots import plot_ecg_rest
from ml4h.runtime_data_defines import get_resting_ecg_hd5_folder
from ml4h.runtime_data_defines import get_resting_ecg_svg_folder
import tensorflow as tf


def display_resting_ecg(sample_id: Union[int, str], folder: Optional[str] = None) -> Union[HTML, SVG]:
  """Retrieve (or render) and display the SVG of the resting ECG.

  Args:
    sample_id: The id of the ECG SVG to retrieve.
    folder: The local or Cloud Storage path under which the files reside.

  Returns:
    An IPython SVG object or a notebook-friendly error.
  """
  if folder is None:
    svg_folder = get_resting_ecg_svg_folder(sample_id)
    hd5_folder = get_resting_ecg_hd5_folder(sample_id)
  else:
    svg_folder = folder
    hd5_folder = folder

  with tempfile.TemporaryDirectory() as tmpdirname:
    # First, see if we already have one rendered.
    sample_svg = str(sample_id) + '.svg'
    local_path = os.path.join(tmpdirname, sample_svg)
    try:
      tf.io.gfile.copy(src=os.path.join(svg_folder, sample_svg), dst=local_path)
      return SVG(filename=local_path)
    except (tf.errors.NotFoundError, tf.errors.PermissionDeniedError) as e:
      pass
    # If not, dynamically render a SVG
    sample_hd5 = str(sample_id) + '.hd5'
    local_path = os.path.join(tmpdirname, sample_hd5)
    try:
      tf.io.gfile.copy(src=os.path.join(hd5_folder, sample_hd5), dst=local_path)
    except (tf.errors.NotFoundError, tf.errors.PermissionDeniedError) as e:
      return HTML(f'''
      <div class="alert alert-block alert-danger">
      <b>Warning:</b> Resting ECG not available for sample {sample_id} in {svg_folder} or {hd5_folder}:
      <hr><p><pre>{e.message}</pre></p>
      Use the <kbd>folder</kbd> parameter to read from a different local directory or Cloud Storage bucket.
      </div>''')

    try:
      # We don't need the resulting SVG, so send it to a temporary directory.
      with tempfile.TemporaryDirectory() as tmpdirname:
        return plot_ecg_rest(tensor_paths=[local_path], rows=[0], out_folder=tmpdirname, is_blind=False)
    except Exception as e:  # pylint: disable=broad-except
      return HTML(f'''
        <div class="alert alert-block alert-danger">
        <b>Warning:</b> Unable to render static plot of resting ECG for sample {sample_id} from {hd5_folder}:
        <hr><p><pre>{e}</pre></p>
        </div>''')


def major_breaks_x_resting_ecg(limits: List[float]) -> np.array:
  """Method to compute breaks for plotnine plots of ECG resting data.

  Args:
    limits: The approximate limits.

  Returns:
    The desired limits.
  """
  step = 0.2
  if limits[0] <= 0:
    min_break = 0.0
    max_break = 2.5
  elif limits[0] <= 2.5:
    min_break = 2.5
    max_break = 5.0
  elif limits[0] <= 5.0:
    min_break = 5.0
    max_break = 7.5
  else:
    min_break = 7.5
    max_break = 10.0
  return np.arange(min_break, max_break + step, step)



# tensor_generators.py
#
# On-the-fly data generation of tensors for training or prediction.
#
# October 2018
# Sam Friedman
# sam@broadinstitute.org

# Python 2/3 friendly
from __future__ import print_function

# Imports
import os
import csv
import math
import h5py
import time
import logging
import traceback
import numpy as np
import pandas as pd
from collections import Counter
from multiprocessing import Process, Queue
from itertools import chain
from typing import List, Dict, Tuple, Set, Optional, Iterator, Callable, Any, Union, Type

import tensorflow as tf
from ml4h.defines import TENSOR_EXT, TensorGeneratorABC
from ml4h.ml4ht_integration.tensor_generator import TensorMapDataLoader
from ml4h.TensorMap import TensorMap

np.set_printoptions(threshold=np.inf)


DEFAULT_VALID_RATIO = 0.2
DEFAULT_TEST_RATIO = 0.1

TENSOR_GENERATOR_TIMEOUT = 64
TENSOR_GENERATOR_MAX_Q_SIZE = 32

# TensorGenerator batch indices
BATCH_INPUT_INDEX, BATCH_OUTPUT_INDEX, BATCH_PATHS_INDEX = 0, 1, 2

Path = str
PathIterator = Iterator[Path]
Batch = Dict[Path, np.ndarray]
BatchFunction = Callable[[Batch, Batch, bool, List[Path], 'kwargs'], Any]


class _ShufflePaths(Iterator):

    def __init__(self, paths: List[Path]):
        self.paths = paths
        np.random.shuffle(self.paths)
        self.idx = 0

    def __next__(self):
        if self.idx >= len(self.paths):
            self.idx = 0
            np.random.shuffle(self.paths)
        path = self.paths[self.idx]
        self.idx += 1
        return path


class _WeightedPaths(Iterator):

    def __init__(self, paths: List[PathIterator], weights: List[float]):
        self.paths = paths
        self.weights = weights
        if len(paths) != len(weights):
            raise ValueError('Weights must be the same length as paths.')

    def __next__(self) -> str:
        return np.random.choice(np.random.choice(self.paths, self.weights))


def pick_generator(paths, weights, mixup, siamese) -> Type[TensorGeneratorABC]:
    try:
        TensorMapDataLoader.can_apply(paths, weights, mixup, siamese)
        return TensorMapDataLoader
    except NotImplementedError as e:
        logging.warning(f"Could not use new data loader because: {repr(e)}. Defaulting to legacy TensorGenerator.")
        return TensorGenerator


class TensorGenerator(TensorGeneratorABC):
    def __init__(
        self, batch_size: int, input_maps: List[TensorMap], output_maps: List[TensorMap],
        paths: Union[List[str], List[List[str]]], num_workers: int, cache_size: float, weights: List[float] = None,
        keep_paths: bool = False, mixup_alpha: float = 0.0, name: str = 'worker', siamese: bool = False,
        augment: bool = False,
    ):
        """
        :param paths: If weights is provided, paths should be a list of path lists the same length as weights
        """
        self.augment = augment
        self.paths = sum(paths) if isinstance(paths[0], list) else paths
        self.run_on_main_thread = num_workers == 0
        self.q = None
        self.stats_q = None
        self._started = False
        self.workers = []
        self.worker_instances = []
        if num_workers == 0:
            num_workers = 1  # The one worker is the main thread
        self.batch_size, self.input_maps, self.output_maps, self.num_workers, self.cache_size, self.weights, self.name, self.keep_paths = \
            batch_size, input_maps, output_maps, num_workers, cache_size, weights, name, keep_paths
        self.true_epochs = 0
        self.stats_string = ""
        if weights is None:
            worker_paths = np.array_split(paths, num_workers)
            self.true_epoch_lens = list(map(len, worker_paths))
            self.path_iters = [_ShufflePaths(p) for p in worker_paths]
        else:
            # split each path list into paths for each worker.
            # E.g. for two workers: [[p1, p2], [p3, p4, p5]] -> [[[p1], [p2]], [[p3, p4], [p5]]
            split_paths = [np.array_split(a, num_workers) for a in paths]
            # Next, each list of paths gets given to each worker. E.g. [[[p1], [p3, p4]], [[p2], [p5]]]
            worker_paths = np.swapaxes(split_paths, 0, 1)
            self.true_epoch_lens = [max(map(len, p)) for p in worker_paths]
            self.path_iters = [_WeightedPaths(p, weights) for p in worker_paths]

        self.batch_function_kwargs = {}
        if mixup_alpha > 0:
            self.batch_function = _mixup_batch
            self.batch_size *= 2
            self.batch_function_kwargs = {'alpha': mixup_alpha}
        elif siamese:
            self.batch_function = _make_batch_siamese
        else:
            self.batch_function = _identity_batch

    def _init_workers(self):
        self.q = Queue(min(self.batch_size, TENSOR_GENERATOR_MAX_Q_SIZE))
        self.stats_q = Queue(len(self.worker_instances))
        self._started = True
        for i, (path_iter, iter_len) in enumerate(zip(self.path_iters, self.true_epoch_lens)):
            name = f'{self.name}_{i}'
            worker_instance = _MultiModalMultiTaskWorker(
                self.q,
                self.stats_q,
                self.num_workers,
                self.input_maps, self.output_maps,
                path_iter, iter_len,
                self.batch_function, self.batch_size, self.keep_paths, self.batch_function_kwargs,
                self.cache_size,
                name,
                self.augment,
            )
            self.worker_instances.append(worker_instance)
            if not self.run_on_main_thread:
                process = Process(
                    target=worker_instance.multiprocessing_worker, name=name,
                    args=(),
                )
                process.start()
                self.workers.append(process)
        logging.info(f"Started {i + 1} {self.name.replace('_', ' ')}s with cache size {self.cache_size/1e9}GB.")

    def set_worker_paths(self, paths: List[Path]):
        """In the single worker case, set the worker's paths."""
        if not self._started:
            self._init_workers()
        if not self.run_on_main_thread:
            raise ValueError('Cannot sort paths of multiprocessing workers. num_workers must be 0.')
        self.worker_instances[0].path_iter.paths = paths

    def __next__(self) -> Tuple[Dict[str, np.ndarray], Dict[str, np.ndarray], Optional[List[str]]]:
        if not self._started:
            self._init_workers()
        if self.stats_q.qsize() == self.num_workers:
            self.aggregate_and_print_stats()
        if self.run_on_main_thread:
            return next(self.worker_instances[0])
        else:
            return self.q.get(TENSOR_GENERATOR_TIMEOUT)

    def aggregate_and_print_stats(self):
        stats = Counter()
        self.true_epochs += 1
        cur_worker = 0
        while self.stats_q.qsize() != 0:
            cur_worker += 1
            worker_stats = self.stats_q.get().copy()
            for k in worker_stats:
                if stats[k] == 0 and cur_worker == 1 and ('_max' in k or '_min' in k):
                    stats[k] = worker_stats[k]
                elif '_max' in k:
                    stats[k] = max(stats[k], worker_stats[k])
                elif '_min' in k:
                    stats[k] = min(stats[k], worker_stats[k])
                else:
                    stats[k] += worker_stats[k]

        all_errors = [
            f'[{error}] - {count:.0f}'
            for error, count in sorted(stats.items(), key=lambda x: x[1], reverse=True) if 'Error' in error
        ]
        if len(all_errors) > 0:
            error_info = f'The following errors were raised:\n\t\t' + '\n\t\t'.join(all_errors)
        else:
            error_info = 'No errors raised.'

        eps = 1e-7
        for tm in self.input_maps + self.output_maps:
            if self.true_epochs != 1:
                break
            if tm.is_categorical() and tm.axes() == 1:
                n = stats[f'{tm.name}_n'] + eps
                self.stats_string = f'{self.stats_string}\nCategorical TensorMap: {tm.name} has {n:.0f} total examples.'
                for channel, index in tm.channel_map.items():
                    examples = stats[f'{tm.name}_index_{index:.0f}']
                    self.stats_string = f'{self.stats_string}\n\tLabel {channel} {examples} examples, {100 * (examples / n):0.2f}% of total.'
            elif tm.is_continuous() and tm.axes() == 1:
                sum_squared = stats[f'{tm.name}_sum_squared']
                n = stats[f'{tm.name}_n'] + eps
                n_sum = stats[f'{tm.name}_sum']
                mean = n_sum / n
                std = np.sqrt((sum_squared/n)-(mean*mean))
                self.stats_string = f'{self.stats_string}\nContinuous TensorMap: {tm.name} has {n:.0f} total examples.\n\tMean: {mean:0.2f}, '
                self.stats_string = f"{self.stats_string}Standard Deviation: {std:0.2f}, Max: {stats[f'{tm.name}_max']:0.2f}, Min: {stats[f'{tm.name}_min']:0.2f}"
            elif tm.is_time_to_event():
                sum_squared = stats[f'{tm.name}_sum_squared']
                n = stats[f'{tm.name}_n'] + eps
                n_sum = stats[f'{tm.name}_sum']
                mean = n_sum / n
                std = np.sqrt((sum_squared/n)-(mean*mean))
                self.stats_string = f"{self.stats_string}\nTime to event TensorMap: {tm.name} Total events: {stats[f'{tm.name}_events']}, "
                self.stats_string = f"{self.stats_string}\n\tMean Follow Up: {mean:0.2f}, Standard Deviation: {std:0.2f}, "
                self.stats_string = f"{self.stats_string}\n\tMax Follow Up: {stats[f'{tm.name}_max']:0.2f}, Min Follow Up: {stats[f'{tm.name}_min']:0.2f}"

        info_string = '\n\t'.join([
            f"Generator looped & shuffled over {sum(self.true_epoch_lens)} paths. Epoch: {self.true_epochs:.0f}",
            f"{stats['Tensors presented']:0.0f} tensors were presented.",
            f"{stats['skipped_paths']} paths were skipped because they previously failed.",
            f"{error_info}",
            f"{self.stats_string}",
        ])
        logging.info(f"\n!!!!>~~~~~~~~~~~~ {self.name} completed true epoch {self.true_epochs} ~~~~~~~~~~~~<!!!!\nAggregated information string:\n\t{info_string}")

    def kill_workers(self):
        if self._started and not self.run_on_main_thread:
            for worker in self.workers:
                worker.terminate()
            logging.info(f'Stopped {len(self.workers)} workers. {self.stats_string}')
        self.workers = []

    def __iter__(self):  # This is so python type annotations recognize TensorGenerator as an iterator
        return self

    def __del__(self):
        self.kill_workers()


class TensorMapArrayCache:
    """
    Caches numpy arrays created by tensor maps up to a maximum number of bytes
    """

    def __init__(self, max_size, input_tms: List[TensorMap], output_tms: List[TensorMap], max_rows: Optional[int] = np.inf):
        input_tms = [tm for tm in input_tms if tm.cacheable]
        output_tms = [tm for tm in output_tms if tm.cacheable]
        self.max_size = max_size
        self.data = {}
        self.row_size = sum(np.zeros(tm.static_shape(), dtype=np.float32).nbytes for tm in set(input_tms + output_tms))
        self.nrows = min(int(max_size / self.row_size), max_rows) if self.row_size else 0
        self.autoencode_names: Dict[str, str] = {}
        for tm in input_tms:
            self.data[tm.input_name()] = np.zeros((self.nrows,) + tm.static_shape(), dtype=np.float32)
        for tm in output_tms:
            if tm in input_tms:  # Useful for autoencoders
                self.autoencode_names[tm.output_name()] = tm.input_name()
            else:
                self.data[tm.output_name()] = np.zeros((self.nrows,) + tm.static_shape(), dtype=np.float32)
        self.files_seen = Counter()  # name -> max position filled in cache
        self.key_to_index = {}  # file_path, name -> position in self.data
        self.hits = 0
        self.failed_paths: Set[str] = set()

    def _fix_key(self, key: Tuple[str, str]) -> Tuple[str, str]:
        file_path, name = key
        return file_path, self.autoencode_names.get(name, name)

    def __setitem__(self, key: Tuple[str, str], value) -> bool:
        """
        :param key: should be a tuple file_path, name
        """
        file_path, name = self._fix_key(key)
        if key in self.key_to_index:  # replace existing value
            self.data[name][self.key_to_index[key]] = value
            return True
        if self.files_seen[name] >= self.nrows:  # cache already full
            return False
        self.key_to_index[key] = self.files_seen[name]
        self.data[name][self.key_to_index[key]] = value
        self.files_seen[name] += 1
        return True

    def __getitem__(self, key: Tuple[str, str]):
        """
        :param key: should be a tuple file_path, name
        """
        file_path, name = self._fix_key(key)
        val = self.data[name][self.key_to_index[file_path, name]]
        self.hits += 1
        return val

    def __contains__(self, key: Tuple[str, str]):
        return self._fix_key(key) in self.key_to_index

    def __len__(self):
        return sum(self.files_seen.values())

    def average_fill(self):
        return np.mean(list(self.files_seen.values()) or [0]) / self.nrows if self.nrows else 0

    def __str__(self):
        hits = f"The cache has had {self.hits} hits."
        fullness = ' - '.join(f"{name} has {count} / {self.nrows} tensors" for name, count in self.files_seen.items())
        return f'{hits} {fullness}.'


class _MultiModalMultiTaskWorker:

    def __init__(
        self,
        q: Queue,
        stats_q: Queue,
        num_workers: int,
        input_maps: List[TensorMap], output_maps: List[TensorMap],
        path_iter: PathIterator, true_epoch_len: int,
        batch_function: BatchFunction, batch_size: int, return_paths: bool, batch_func_kwargs: Dict,
        cache_size: float,
        name: str,
        augment: bool,
    ):
        self.q = q
        self.stats_q = stats_q
        self.num_workers = num_workers
        self.input_maps = input_maps
        self.output_maps = output_maps
        self.path_iter = path_iter
        self.true_epoch_len = true_epoch_len
        self.batch_function = batch_function
        self.batch_size = batch_size
        self.return_paths = return_paths
        self.batch_func_kwargs = batch_func_kwargs
        self.cache_size = cache_size
        self.name = name
        self.augment = augment

        self.stats = Counter()
        self.epoch_stats = Counter()
        self.start = time.time()
        self.paths_in_batch = []

        self.in_batch = {tm.input_name(): np.zeros((batch_size,) + tm.static_shape()) for tm in input_maps}
        self.out_batch = {tm.output_name(): np.zeros((batch_size,) + tm.static_shape()) for tm in output_maps}

        self.cache = TensorMapArrayCache(cache_size, input_maps, output_maps, true_epoch_len)
        self.dependents = {}
        self.idx = 0

    def _handle_tm(self, tm: TensorMap, is_input: bool, path: Path) -> h5py.File:
        name = tm.input_name() if is_input else tm.output_name()
        batch = self.in_batch if is_input else self.out_batch
        idx = self.stats['batch_index']

        if tm in self.dependents:
            batch[name][idx] = self.dependents[tm]
            if tm.cacheable:
                self.cache[path, name] = self.dependents[tm]
            self._collect_stats(tm, self.dependents[tm])
            return self.hd5
        if (path, name) in self.cache:
            batch[name][idx] = self.cache[path, name]
            return self.hd5
        if self.hd5 is None:  # Don't open hd5 if everything is in the self.cache
            self.hd5 = h5py.File(path, 'r')
        tensor = tm.postprocess_tensor(tm.tensor_from_file(tm, self.hd5, self.dependents), augment=self.augment, hd5=self.hd5)
        slices = tuple(slice(min(tm.static_shape()[i], tensor.shape[i])) for i in range(len(tensor.shape)))
        batch[name][(idx,)+slices] = tensor[slices]
        if tm.cacheable:
            self.cache[path, name] = batch[name][idx]
        self._collect_stats(tm, tensor)
        return self.hd5

    def _collect_stats(self, tm, tensor):
        if tm.is_time_to_event():
            self.epoch_stats[f'{tm.name}_events'] += tensor[0]
            self._collect_continuous_stats(tm, tensor[1])
        if tm.is_categorical() and tm.axes() == 1:
            self.epoch_stats[f'{tm.name}_index_{np.argmax(tensor):.0f}'] += 1
        if tm.is_continuous() and tm.axes() == 1:
            self._collect_continuous_stats(tm, tm.rescale(tensor)[0])
        self.epoch_stats[f'{tm.name}_n'] += 1

    def _collect_continuous_stats(self, tm, rescaled):
        if 0.0 == self.epoch_stats[f'{tm.name}_max'] == self.epoch_stats[f'{tm.name}_min']:
            self.epoch_stats[f'{tm.name}_max'] = rescaled
            self.epoch_stats[f'{tm.name}_min'] = rescaled
        self.epoch_stats[f'{tm.name}_max'] = max(rescaled, self.epoch_stats[f'{tm.name}_max'])
        self.epoch_stats[f'{tm.name}_min'] = min(rescaled, self.epoch_stats[f'{tm.name}_min'])
        self.epoch_stats[f'{tm.name}_sum'] += rescaled
        self.epoch_stats[f'{tm.name}_sum_squared'] += rescaled * rescaled

    def _handle_tensor_path(self, path: Path) -> None:
        hd5 = None
        if path in self.cache.failed_paths:
            self.epoch_stats['skipped_paths'] += 1
            return
        try:
            self.dependents = {}
            self.hd5 = None
            for tm in self.input_maps:
                hd5 = self._handle_tm(tm, True, path)
            for tm in self.output_maps:
                hd5 = self._handle_tm(tm, False, path)
            self.paths_in_batch.append(path)
            self.stats['Tensors presented'] += 1
            self.epoch_stats['Tensors presented'] += 1
            self.stats['batch_index'] += 1
        except (IndexError, KeyError, ValueError, OSError, RuntimeError) as e:
            error_name = type(e).__name__
            self.stats[f"{error_name} while attempting to generate tensor:\n{traceback.format_exc()}\n"] += 1
            self.epoch_stats[f"{error_name}: {e}"] += 1
            self.cache.failed_paths.add(path)
            _log_first_error(self.stats, path)
        finally:
            if hd5 is not None:
                hd5.close()

    def _on_epoch_end(self):
        self.stats['epochs'] += 1
        self.epoch_stats['epochs'] = self.stats['epochs']
        while self.stats_q.qsize() == self.num_workers:
            continue
        self.stats_q.put(self.epoch_stats)
        if self.stats['Tensors presented'] == 0:
            logging.error(f"Completed an epoch but did not find any tensors to yield")
        if 'test' in self.name:
            logging.warning(f'Test worker {self.name} completed a full epoch. Test results may be double counting samples.')
        self.start = time.time()
        self.epoch_stats = Counter()

    def multiprocessing_worker(self):
        for i, path in enumerate(self.path_iter):
            self._handle_tensor_path(path)
            if self.stats['batch_index'] == self.batch_size:

                out = self.batch_function(self.in_batch, self.out_batch, self.return_paths, self.paths_in_batch, **self.batch_func_kwargs)
                self.q.put(out)
                self.paths_in_batch = []
                self.stats['batch_index'] = 0
                self.in_batch = {tm.input_name(): np.zeros((self.batch_size,) + tm.static_shape()) for tm in self.input_maps}
                self.out_batch = {tm.output_name(): np.zeros((self.batch_size,) + tm.static_shape()) for tm in self.output_maps}
            if i > 0 and i % self.true_epoch_len == 0:
                self._on_epoch_end()

    def __next__(self):
        while self.stats['batch_index'] < self.batch_size:
            path = next(self.path_iter)
            self._handle_tensor_path(path)
            if self.idx > 0 and self.idx % self.true_epoch_len == 0:
                self._on_epoch_end()
            self.idx += 1
        self.stats['batch_index'] = 0
        out = self.batch_function(self.in_batch, self.out_batch, self.return_paths, self.paths_in_batch, **self.batch_func_kwargs)
        self.paths_in_batch = []
        return out


def big_batch_from_minibatch_generator(generator: TensorGenerator, minibatches: int, keep_paths: bool = True):
    """Collect minibatches into bigger batches

    Returns a dicts of numpy arrays like the same kind as generator but with more examples.

    Arguments:
        generator: TensorGenerator of minibatches
        minibatches: number of times to call generator and collect a minibatch

    Returns:
        A tuple of dicts mapping tensor names to big batches of numpy arrays mapping.
    """
    first_batch = next(generator)
    saved_tensors = {}
    batch_size = None
    for key, batch_array in chain(first_batch[BATCH_INPUT_INDEX].items(), first_batch[BATCH_OUTPUT_INDEX].items()):
        shape = (batch_array.shape[0] * minibatches,) + batch_array.shape[1:]
        saved_tensors[key] = np.zeros(shape)
        batch_size = batch_array.shape[0]
        saved_tensors[key][:batch_size] = batch_array

    if keep_paths:
        paths = first_batch[BATCH_PATHS_INDEX]

    input_tensors, output_tensors = list(first_batch[BATCH_INPUT_INDEX]), list(first_batch[BATCH_OUTPUT_INDEX])
    for i in range(1, minibatches):
        logging.debug(f'big_batch_from_minibatch {100 * i / minibatches:.2f}% done.')
        next_batch = next(generator)
        s, t = i * batch_size, (i + 1) * batch_size
        for key in input_tensors:
            saved_tensors[key][s:t] = next_batch[BATCH_INPUT_INDEX][key]
        for key in output_tensors:
            saved_tensors[key][s:t] = next_batch[BATCH_OUTPUT_INDEX][key]
        if keep_paths:
            paths.extend(next_batch[BATCH_PATHS_INDEX])

    for key, array in saved_tensors.items():
        logging.info(f"Made a big batch of tensors with key:{key} and shape:{array.shape}.")
    inputs = {key: saved_tensors[key] for key in input_tensors}
    outputs = {key: saved_tensors[key] for key in output_tensors}
    if keep_paths:
        return inputs, outputs, paths
    else:
        return inputs, outputs


def _get_train_valid_test_discard_ratios(
        valid_ratio: float,
        test_ratio: float,
        train_csv: str,
        valid_csv: str,
        test_csv: str,
) -> Tuple[int, int, int, int]:

    if valid_csv is not None:
        valid_ratio = 0
    if test_csv is not None:
        test_ratio = 0
    if train_csv is not None:
        train_ratio = 0
        discard_ratio = 1.0 - valid_ratio - test_ratio
    else:
        train_ratio = 1.0 - valid_ratio - test_ratio
        discard_ratio = 0

    if not math.isclose(train_ratio + valid_ratio + test_ratio + discard_ratio, 1.0):
        raise ValueError(f'ratios do not sum to 1, train/valid/test/discard = {train_ratio}/{valid_ratio}/{test_ratio}/{discard_ratio}')
    logging.debug(f'train/valid/test/discard ratios: {train_ratio}/{valid_ratio}/{test_ratio}/{discard_ratio}')

    return train_ratio, valid_ratio, test_ratio, discard_ratio


def _sample_csv_to_set(sample_csv: Optional[str] = None) -> Union[None, Set[str]]:
    if sample_csv is None:
        return None

    # Read CSV to dataframe and assume no header
    df = pd.read_csv(sample_csv, header=None)

    # If first row and column is castable to int, there is no header
    try:
        int(df.iloc[0].values[0])
    # If fails, must be header; overwrite column name with first row and remove first row
    except ValueError:
        df.columns = df.iloc[0]
        df = df[1:]

    # Declare set of possible MRN column names
    possible_mrn_col_names = {"sampleid", "medrecn", "mrn", "patient_id"}

    # Find intersection between CSV columns and possible MRN column names
    matches = set(df.columns).intersection(possible_mrn_col_names)

    # If no matches, assume the first column is MRN
    if not matches:
        mrn_col_name = df.columns[0]
    else:
         # Get first string from set of matches to use as column name
        mrn_col_name = next(iter(matches))

    if len(matches) > 1:
        logging.warning(
            f"{sample_csv} has more than one potential column for MRNs. Inferring most likely column name, but recommend explicitly setting MRN column name.",
        )

    # Isolate this column from the dataframe, and cast to strings
    sample_ids = df[mrn_col_name].apply(str)

    return set(sample_ids)


def get_train_valid_test_paths(
        tensors: str,
        sample_csv: str,
        valid_ratio: float,
        test_ratio: float,
        train_csv: str,
        valid_csv: str,
        test_csv: str,
) -> Tuple[List[str], List[str], List[str]]:
    """
    Return 3 disjoint lists of tensor paths.

    The paths are split in training, validation, and testing lists.
    If no arguments are given, paths are split into train/valid/test in the ratio 0.7/0.2/0.1.
    Otherwise, at least 2 arguments are required to specify train/valid/test sets.

    :param tensors: path to directory containing tensors
    :param sample_csv: path to csv containing sample ids, only consider sample ids for splitting
                       into train/valid/test sets if they appear in sample_csv
    :param valid_ratio: rate of tensors in validation list, mutually exclusive with valid_csv
    :param test_ratio: rate of tensors in testing list, mutually exclusive with test_csv
    :param train_csv: path to csv containing sample ids to reserve for training list
    :param valid_csv: path to csv containing sample ids to reserve for validation list, mutually exclusive with valid_ratio
    :param test_csv: path to csv containing sample ids to reserve for testing list, mutually exclusive with test_ratio

    :return: tuple of 3 lists of hd5 tensor file paths
    """
    train_paths = []
    valid_paths = []
    test_paths = []
    discard_paths = []

    train_ratio, valid_ratio, test_ratio, discard_ratio = _get_train_valid_test_discard_ratios(
        valid_ratio=valid_ratio,
        test_ratio=test_ratio,
        train_csv=train_csv,
        valid_csv=valid_csv,
        test_csv=test_csv,
    )

    choices = {
        'train': (train_paths, train_ratio),
        'valid': (valid_paths, valid_ratio),
        'test': (test_paths, test_ratio),
        'discard': (discard_paths, discard_ratio),
    }

    # parse csv's to disjoint sets, None if csv was None
    sample_set = _sample_csv_to_set(sample_csv)

    train_set = _sample_csv_to_set(train_csv)
    valid_set = _sample_csv_to_set(valid_csv)
    test_set = _sample_csv_to_set(test_csv)

    if train_set is not None and valid_set is not None and not train_set.isdisjoint(valid_set):
        raise ValueError('train and validation samples overlap')
    if train_set is not None and test_set is not None and not train_set.isdisjoint(test_set):
        raise ValueError('train and test samples overlap')
    if valid_set is not None and test_set is not None and not valid_set.isdisjoint(test_set):
        raise ValueError('validation and test samples overlap')

    # find tensors and split them among train/valid/test
    for root, dirs, files in os.walk(tensors):
        for name in files:
            path = os.path.join(root, name)
            split = os.path.splitext(name)
            sample_id = split[0]

            if split[-1].lower() != TENSOR_EXT:
                continue
            elif sample_set is not None and sample_id not in sample_set:
                continue
            elif train_set is not None and sample_id in train_set:
                train_paths.append(path)
            elif valid_set is not None and sample_id in valid_set:
                valid_paths.append(path)
            elif test_set is not None and sample_id in test_set:
                test_paths.append(path)
            else:
                choice = np.random.choice([k for k in choices], p=[choices[k][1] for k in choices])
                choices[choice][0].append(path)

    logging.info(f'Found {len(train_paths)} train, {len(valid_paths)} validation, and {len(test_paths)} testing tensors at: {tensors}')
    logging.debug(f'Discarded {len(discard_paths)} tensors due to given ratios')
    if len(train_paths) == 0 or len(valid_paths) == 0 or len(test_paths) == 0:
        raise ValueError(
            f'Not enough tensors at {tensors}\n'
            f'Found {len(train_paths)} training, {len(valid_paths)} validation, and {len(test_paths)} testing tensors\n'
            f'Discarded {len(discard_paths)} tensors',
        )

    return train_paths, valid_paths, test_paths


def get_train_valid_test_paths_split_by_csvs(
        tensors: str,
        balance_csvs: List[str],
        sample_csv: str,
        valid_ratio: float,
        test_ratio: float,
        train_csv: str,
        valid_csv: str,
        test_csv: str,
) -> Tuple[List[List[str]], List[List[str]], List[List[str]]]:
    stats = Counter()
    sample2group = {}
    for i, b_csv in enumerate(balance_csvs):
        lol = list(csv.reader(open(b_csv, 'r'), delimiter=','))
        logging.info(f"Class Balance CSV Header: {list(enumerate(lol[0]))}")

        for row in lol[1:]:
            sample_id = row[0]
            sample2group[sample_id] = i+1  # group 0 means background class
            stats['group_'+str(i+1)] += 1
    logging.info(f"Balancing with CSVs of Sample IDs stats: {stats}")

    train_paths = [[] for _ in range(len(balance_csvs)+1)]
    valid_paths = [[] for _ in range(len(balance_csvs)+1)]
    test_paths = [[] for _ in range(len(balance_csvs)+1)]

    _train, _valid, _test = get_train_valid_test_paths(
        tensors=tensors,
        sample_csv=sample_csv,
        valid_ratio=valid_ratio,
        test_ratio=test_ratio,
        train_csv=train_csv,
        valid_csv=valid_csv,
        test_csv=test_csv,
    )

    for paths, split_list in [(_train, train_paths), (_valid, valid_paths), (_test, test_paths)]:
        for path in paths:
            split = os.path.splitext(os.path.basename(path))
            sample_id = split[0]

            group = 0
            if sample_id in sample2group:
                group = sample2group[sample_id]
            split_list[group].append(path)

    for i in range(len(train_paths)):
        if len(train_paths[i]) == 0 or len(valid_paths[i]) == 0 or len(test_paths[i]) == 0:
            my_error = f"Not enough tensors at {tensors}\nGot {len(train_paths[i])} train {len(valid_paths[i])} valid and {len(test_paths[i])} test."
            raise ValueError(my_error)
        if i == 0:
            logging.info(f"Found {len(train_paths[i])} train {len(valid_paths[i])} valid and {len(test_paths[i])} test tensors outside the CSVs.")
        else:
            logging.info(f"CSV:{balance_csvs[i-1]}\nhas: {len(train_paths[i])} train, {len(valid_paths[i])} valid, {len(test_paths[i])} test tensors.")
    return train_paths, valid_paths, test_paths


def test_train_valid_tensor_generators(
    tensor_maps_in: List[TensorMap],
    tensor_maps_out: List[TensorMap],
    tensor_maps_protected: List[TensorMap],
    tensors: str,
    batch_size: int,
    num_workers: int,
    training_steps: int,
    validation_steps: int,
    cache_size: float,
    balance_csvs: List[str],
    keep_paths: bool = False,
    keep_paths_test: bool = True,
    mixup_alpha: float = -1.0,
    sample_csv: str = None,
    valid_ratio: float = None,
    test_ratio: float = None,
    train_csv: str = None,
    valid_csv: str = None,
    test_csv: str = None,
    siamese: bool = False,
    wrap_with_tf_dataset: bool = False,
    **kwargs
) -> Tuple[TensorGeneratorABC, TensorGeneratorABC, TensorGeneratorABC]:
    """ Get 3 tensor generator functions for training, validation and testing data.

    :param tensor_maps_in: list of TensorMaps that are input names to a model
    :param tensor_maps_out: list of TensorMaps that are output from a model
    :param tensor_maps_protected: list of TensorMaps that are sensitive to bias from a model
                                    only added to the test set
    :param tensors: directory containing tensors
    :param batch_size: number of examples in each mini-batch
    :param num_workers: number of processes spun off for training and testing. Validation uses half as many workers
    :param training_steps: Number of training batches that define a fake "epoch"
    :param validation_steps: Number of validation batches to create after each fake "epoch"
    :param cache_size: size in bytes of maximum cache for EACH worker
    :param balance_csvs: if not empty, generator will provide batches balanced amongst the Sample ID in these CSVs.
    :param keep_paths: also return the list of tensor files loaded for training and validation tensors
    :param keep_paths_test:  also return the list of tensor files loaded for testing tensors
    :param mixup_alpha: If positive, mixup batches and use this value as shape parameter alpha
    :param sample_csv: CSV file of sample ids, sample ids are considered for train/valid/test only if it is in sample_csv
    :param valid_ratio: rate of tensors to use for validation, mutually exclusive with valid_csv
    :param test_ratio: rate of tensors to use for testing, mutually exclusive with test_csv
    :param train_csv: CSV file of sample ids to use for training
    :param valid_csv: CSV file of sample ids to use for validation, mutually exclusive with valid_ratio
    :param test_csv: CSV file of sample ids to use for testing, mutually exclusive with test_ratio
    :param siamese: if True generate input for a siamese model i.e. a left and right input tensors for every input TensorMap
    :param wrap_with_tf_dataset: if True will return tf.dataset objects for the 3 generators
    :return: A tuple of three generators. Each yields a Tuple of dictionaries of input and output numpy arrays for training, validation and testing.
    """
    generate_train, generate_valid, generate_test = None, None, None
    if len(balance_csvs) > 0:
        train_paths, valid_paths, test_paths = get_train_valid_test_paths_split_by_csvs(
            tensors=tensors,
            balance_csvs=balance_csvs,
            sample_csv=sample_csv,
            valid_ratio=valid_ratio,
            test_ratio=test_ratio,
            train_csv=train_csv,
            valid_csv=valid_csv,
            test_csv=test_csv,
        )
        weights = [1.0/(len(balance_csvs)+1) for _ in range(len(balance_csvs)+1)]
    else:
        train_paths, valid_paths, test_paths = get_train_valid_test_paths(
            tensors=tensors,
            sample_csv=sample_csv,
            valid_ratio=valid_ratio,
            test_ratio=test_ratio,
            train_csv=train_csv,
            valid_csv=valid_csv,
            test_csv=test_csv,
        )
        weights = None

    num_train_workers = int(training_steps / (training_steps + validation_steps) * num_workers) or (1 if num_workers else 0)
    num_valid_workers = int(validation_steps / (training_steps + validation_steps) * num_workers) or (1 if num_workers else 0)

    generator_class = pick_generator(train_paths, weights, mixup_alpha, siamese)

    generate_train = generator_class(
        batch_size=batch_size, input_maps=tensor_maps_in, output_maps=tensor_maps_out,
        paths=train_paths, num_workers=num_train_workers, cache_size=cache_size, weights=weights,
        keep_paths=keep_paths, mixup_alpha=mixup_alpha, name='train_worker', siamese=siamese, augment=True,
    )
    generate_valid = generator_class(
        batch_size=batch_size, input_maps=tensor_maps_in, output_maps=tensor_maps_out,
        paths=valid_paths, num_workers=num_valid_workers, cache_size=cache_size, weights=weights,
        keep_paths=keep_paths, mixup_alpha=0, name='validation_worker', siamese=siamese, augment=False,
    )
    generate_test = generator_class(
        batch_size=batch_size, input_maps=tensor_maps_in, output_maps=tensor_maps_out + tensor_maps_protected,
        paths=test_paths, num_workers=num_train_workers, cache_size=0, weights=weights,
        keep_paths=keep_paths or keep_paths_test, mixup_alpha=0, name='test_worker', siamese=siamese, augment=False,
    )
    if wrap_with_tf_dataset:
        in_shapes = {tm.input_name(): (batch_size,) + tm.static_shape() for tm in tensor_maps_in}
        out_shapes = {tm.output_name(): (batch_size,) + tm.static_shape() for tm in tensor_maps_out}
        train_dataset = tf.data.Dataset.from_generator(
            generate_train,
            output_types=({k: tf.float32 for k in in_shapes}, {k: tf.float32 for k in out_shapes}),
            output_shapes=(in_shapes, out_shapes),
        )
        valid_dataset = tf.data.Dataset.from_generator(
            generate_valid,
            output_types=({k: tf.float32 for k in in_shapes}, {k: tf.float32 for k in out_shapes}),
            output_shapes=(in_shapes, out_shapes),
        )
        test_dataset = tf.data.Dataset.from_generator(
            generate_test,
            output_types=({k: tf.float32 for k in in_shapes}, {k: tf.float32 for k in out_shapes}),
            output_shapes=(in_shapes, out_shapes),
        )
        return train_dataset, valid_dataset, test_dataset
    else:
        return generate_train, generate_valid, generate_test


def _log_first_error(stats: Counter, tensor_path: str):
    for k in stats:
        if 'Error' in k and stats[k] == 1:
            stats[k] += 1  # Increment so we only see these messages once
            logging.debug(f"At tensor path: {tensor_path}")
            logging.debug(f"Got first error: {k}")


def _identity_batch(in_batch: Batch, out_batch: Batch, return_paths: bool, paths: List[Path]):
    return (in_batch, out_batch, paths) if return_paths else (in_batch, out_batch)


def _mixup_batch(in_batch: Batch, out_batch: Batch, return_paths: bool, paths: List[Path], alpha: float = 1.0, permute_first: bool = False):
    full_batch = in_batch.values().__iter__().__next__().shape[0]
    half_batch = full_batch // 2

    if permute_first:
        permuted = np.random.permutation(full_batch)
        for k in in_batch:
            in_batch[k] = in_batch[k][permuted, ...]
        for k in out_batch:
            out_batch[k] = out_batch[k][permuted, ...]

    mixed_ins = {k: np.zeros((half_batch,) + in_batch[k].shape[1:]) for k in in_batch}
    mixed_outs = {k: np.zeros((half_batch,) + out_batch[k].shape[1:]) for k in out_batch}
    for i in range(half_batch):
        weight0 = np.random.beta(alpha, alpha)
        weight1 = 1 - weight0
        for k in in_batch:
            mixed_ins[k][i] = (in_batch[k][i, ...] * weight0) + (in_batch[k][half_batch + i, ...] * weight1)
        for k in out_batch:
            mixed_outs[k][i] = (out_batch[k][i, ...] * weight0) + (out_batch[k][half_batch + i, ...] * weight1)

    return _identity_batch(mixed_ins, mixed_outs, return_paths, paths[:half_batch])


def _make_batch_siamese(in_batch: Batch, out_batch: Batch, return_paths: bool, paths: List[Path]):
    full_batch = in_batch.values().__iter__().__next__().shape[0]
    half_batch = full_batch // 2

    siamese_in = {k+'_left': np.zeros((half_batch,) + in_batch[k].shape[1:]) for k in in_batch}
    siamese_in.update({k+'_right': np.zeros((half_batch,) + in_batch[k].shape[1:]) for k in in_batch})
    siamese_out = {'output_siamese': np.zeros((half_batch, 1))}

    for i in range(half_batch):
        for k in in_batch:
            siamese_in[k+'_left'][i] = in_batch[k][i, ...]
            siamese_in[k+'_right'][i] = in_batch[k][half_batch + i, ...]
        random_task_key = np.random.choice(list(out_batch.keys()))
        siamese_out['output_siamese'][i] = 0 if np.array_equal(out_batch[random_task_key][i], out_batch[random_task_key][i+half_batch]) else 1

    return _identity_batch(siamese_in, siamese_out, return_paths, paths)


def _weighted_batch(in_batch: Batch, out_batch: Batch, return_paths: bool, paths: List[Path]):
    return (in_batch, out_batch, paths) if return_paths else (in_batch, out_batch)



import logging
from typing import Dict, List, Tuple, Sequence

import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Dense, Flatten, Reshape, LayerNormalization, DepthwiseConv2D, concatenate, Concatenate, Add

from ml4h.models.Block import Block
from ml4h.TensorMap import TensorMap
from ml4h.models.basic_blocks import DenseBlock
from ml4h.models.layer_wrappers import _upsampler, _activation_layer, _regularization_layer, _normalization_layer
from ml4h.models.layer_wrappers import _conv_layer_from_kind_and_dimension, _pool_layers_from_kind_and_dimension, _one_by_n_kernel

Tensor = tf.Tensor


class ConvEncoderBlock(Block):
    def __init__(
            self,
            *,
            tensor_map: TensorMap,
            dense_blocks: List[int] = [32, 32, 32],
            dense_layers: List[int] = [32],
            dense_normalize: str = None,
            dense_regularize: str = None,
            dense_regularize_rate: float = 0.0,
            conv_layers: List[int] = [32],
            conv_type: str = 'conv',
            conv_width: List[int] = [71],
            conv_x: List[int] = [3],
            conv_y: List[int] = [3],
            conv_z: List[int] = [3],
            block_size: int = 3,
            activation: str = 'swish',
            conv_normalize: str = None,
            conv_regularize: str = None,
            conv_regularize_rate: float = 0.0,
            conv_dilate: bool = False,
            pool_type: str = 'max',
            pool_x: int = 2,
            pool_y: int = 2,
            pool_z: int = 1,
            **kwargs,
    ):
        self.tensor_map = tensor_map
        if not self.can_apply():
            return
        dimension = self.tensor_map.axes()

        # list of filter dimensions should match the total number of convolutional layers
        x_filters = _repeat_dimension(conv_width if dimension == 2 else conv_x, len(conv_layers)+len(dense_blocks))
        y_filters = _repeat_dimension(conv_y, len(conv_layers)+len(dense_blocks))
        z_filters = _repeat_dimension(conv_z, len(conv_layers)+len(dense_blocks))

        #self.preprocess_block = PreprocessBlock(['rotate'], [0.3])
        self.res_block = Residual(
            dimension=dimension, filters_per_conv=conv_layers, conv_layer_type=conv_type, conv_x=x_filters[:len(conv_layers)],
            conv_y=y_filters[:len(conv_layers)], conv_z=z_filters[:len(conv_layers)], activation=activation, normalization=conv_normalize,
            regularization=conv_regularize, regularization_rate=conv_regularize_rate, dilate=conv_dilate,
        )

        self.dense_blocks = [
            DenseConvolutional(
                dimension=dimension, conv_layer_type=conv_type, filters=filters, conv_x=[x] * block_size, conv_y=[y] * block_size,
                conv_z=[z]*block_size, block_size=block_size, activation=activation, normalization=conv_normalize,
                regularization=conv_regularize, regularization_rate=conv_regularize_rate,
            ) for filters, x, y, z in zip(dense_blocks, x_filters[len(conv_layers):], y_filters[len(conv_layers):], z_filters[len(conv_layers):])
        ]
        self.pools = _pool_layers_from_kind_and_dimension(dimension, pool_type, len(dense_blocks) + 1, pool_x, pool_y, pool_z)
        self.fully_connected = DenseBlock(
            widths=dense_layers,
            activation=activation,
            normalization=dense_normalize,
            regularization=dense_regularize,
            regularization_rate=dense_regularize_rate,
            name=self.tensor_map.embed_name(),
        ) if dense_layers else None

    def can_apply(self):
        return self.tensor_map.axes() > 1

    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        if not self.can_apply():
            return x
        #x = self.preprocess_block(x)  # TODO: upgrade to tensorflow 2.3+
        x = self.res_block(x)
        intermediates[self.tensor_map].append(x)
        for i, (dense_block, pool) in enumerate(zip(self.dense_blocks, self.pools)):
            x = pool(x)
            x = dense_block(x)
            intermediates[self.tensor_map].append(x)
        if self.fully_connected:
            x = Flatten()(x)
            x = self.fully_connected(x, intermediates)
            intermediates[self.tensor_map].append(x)
        return x


class ConvDecoderBlock(Block):
    def __init__(
            self,
            *,
            tensor_map: TensorMap,
            dense_blocks: List[int] = [32, 32, 32],
            conv_type: str = 'conv',
            conv_width: List[int] = [71],
            conv_x: List[int] = [3],
            conv_y: List[int] = [3],
            conv_z: List[int] = [3],
            block_size: int = 3,
            activation: str = 'swish',
            conv_normalize: str = None,
            conv_regularize: str = None,
            conv_regularize_rate: float = 0.0,
            pool_x: int = 2,
            pool_y: int = 2,
            pool_z: int = 1,
            u_connect_parents: List[TensorMap] = None,
            **kwargs,
    ):
        self.tensor_map = tensor_map
        if not self.can_apply():
            return
        dimension = tensor_map.axes()
        x_filters = _repeat_dimension(conv_width if dimension == 2 else conv_x, len(dense_blocks))
        y_filters = _repeat_dimension(conv_y, len(dense_blocks))
        z_filters = _repeat_dimension(conv_z, len(dense_blocks))
        self.dense_conv_blocks = [
            DenseConvolutional(
                dimension=tensor_map.axes(), conv_layer_type=conv_type, filters=filters, conv_x=[x] * block_size,
                conv_y=[y]*block_size, conv_z=[z]*block_size, block_size=block_size, activation=activation, normalization=conv_normalize,
                regularization=conv_regularize, regularization_rate=conv_regularize_rate,
            )
            for filters, x, y, z in zip(dense_blocks, x_filters, y_filters, z_filters)
        ]
        conv_layer, _ = _conv_layer_from_kind_and_dimension(dimension, 'conv', conv_x, conv_y, conv_z)
        self.conv_label = conv_layer(tensor_map.shape[-1], _one_by_n_kernel(dimension), activation=tensor_map.activation, name=tensor_map.output_name())
        self.upsamples = [_upsampler(dimension, pool_x, pool_y, pool_z) for _ in range(len(dense_blocks) + 1)]
        self.u_connect_parents = u_connect_parents or []
        self.start_shape = _start_shape_before_pooling(
            num_upsamples=len(dense_blocks), output_shape=tensor_map.shape,
            upsample_rates=[pool_x, pool_y, pool_z], channels=dense_blocks[-1],
        )
        self.reshape = FlatToStructure(output_shape=self.start_shape, activation=activation, normalization=conv_normalize)
        logging.info(f'Built a decoder with: {len(self.dense_conv_blocks)} and reshape {self.start_shape}')

    def can_apply(self):
        return self.tensor_map.axes() > 1

    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        if not self.can_apply():
            return x
        if x.shape != self.start_shape:
            x = self.reshape(x)
        for i, (dense_block, upsample) in enumerate(zip(self.dense_conv_blocks, self.upsamples)):
            intermediate = [intermediates[tm][len(self.upsamples)-(i+1)] for tm in self.u_connect_parents]
            x = concatenate(intermediate + [x]) if intermediate else x
            x = upsample(x)
            x = dense_block(x)
        intermediate = [intermediates[tm][0] for tm in self.u_connect_parents]
        x = concatenate(intermediate + [x]) if intermediate else x
        return self.conv_label(x)


class ResidualBlock(Block):
    def __init__(
            self,
            *,
            tensor_map: TensorMap,
            conv_layers: List[int],
            conv_type: str,
            conv_width: List[int],
            conv_x: List[int],
            conv_y: List[int],
            conv_z: List[int],
            activation: str,
            conv_normalize: str,
            conv_regularize: str,
            conv_regularize_rate: float,
            conv_dilate: bool,
            **kwargs,
    ):
        self.tensor_map = tensor_map
        if not self.can_apply():
            return
        block_size = len(conv_layers)
        x_filters, y_filters, z_filters = _get_xyz_filters(block_size, conv_x if self.tensor_map.axes() > 2 else conv_width, conv_y, conv_z)
        conv_layer, kernels = _conv_layer_from_kind_and_dimension(self.tensor_map.axes(), conv_type, x_filters, y_filters, z_filters)
        self.conv_layers = []
        for i, (num_filters, kernel) in enumerate(zip(conv_layers, kernels)):
            if isinstance(conv_layer, DepthwiseConv2D):
                self.conv_layers.append(conv_layer(kernel_size=kernel, padding='same', dilation_rate=2 ** i if conv_dilate else 1))
            else:
                self.conv_layers.append(conv_layer(filters=num_filters, kernel_size=kernel, padding='same', dilation_rate=2**i if conv_dilate else 1))

        self.activations = [_activation_layer(activation) for _ in range(block_size)]
        self.normalizations = [_normalization_layer(conv_normalize) for _ in range(block_size)]
        self.regularizations = [_regularization_layer(self.tensor_map.axes(), conv_regularize, conv_regularize_rate) for _ in range(block_size)]
        residual_conv_layer, _ = _conv_layer_from_kind_and_dimension(self.tensor_map.axes(), 'conv', x_filters, y_filters, z_filters)
        self.residual_convs = [residual_conv_layer(filters=conv_layers[0], kernel_size=_one_by_n_kernel(self.tensor_map.axes())) for _ in range(block_size - 1)]
        logging.info(f'Residual Block Convolutional Layers (num_filters, kernel_size): {list(zip(conv_layers, kernels))}')

    def can_apply(self):
        return self.tensor_map.axes() > 1

    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        if not self.can_apply():
            return x
        previous = x
        for convolve, activate, normalize, regularize, one_by_n_convolve in zip(
                self.conv_layers, self.activations, self.normalizations, self.regularizations, [None] + self.residual_convs,
        ):
            x = regularize(normalize(activate(convolve(x))))
            if one_by_n_convolve is not None:  # Do not residual add the input
                x = Add()([one_by_n_convolve(x), previous])
            previous = x
        intermediates[self.tensor_map].append(x)
        return x


class PoolBlock(Block):
    def __init__(
            self,
            *,
            tensor_map: TensorMap,
            pool_type: str,
            pool_x: int,
            pool_y: int,
            pool_z: int,
            **kwargs,
    ):
        self.tensor_map = tensor_map
        if not self.can_apply():
            return
        dimension = self.tensor_map.axes()
        self.pool = _pool_layers_from_kind_and_dimension(dimension, pool_type, 1, pool_x, pool_y, pool_z)[0]

    def can_apply(self):
        return self.tensor_map.axes() > 1

    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        if not self.can_apply():
            return x
        x = self.pool(x)
        intermediates[self.tensor_map].append(x)
        return x


class ConvDown(Block):
    def __init__(
            self,
            *,
            tensor_map: TensorMap,
            conv_layers: List[int],
            conv_type: str,
            conv_width: List[int],
            conv_x: List[int],
            conv_y: List[int],
            conv_z: List[int],
            activation: str,
            conv_strides: int,
            conv_dilate: bool,
            conv_normalize: str,
            conv_regularize: str,
            conv_regularize_rate: float,
            conv_without_bias: bool,
            conv_bias_initializer: str,
            conv_kernel_initializer: str,
            dense_layers: List[int],
            dense_normalize: str,
            dense_regularize: str,
            dense_regularize_rate: float,
            **kwargs,
    ):
        self.tensor_map = tensor_map
        if not self.can_apply():
            return
        block_size = len(conv_layers)
        x_filters, y_filters, z_filters = _get_xyz_filters(block_size, conv_x if self.tensor_map.axes() > 2 else conv_width, conv_y, conv_z)
        conv_layer, kernels = _conv_layer_from_kind_and_dimension(self.tensor_map.axes(), conv_type, x_filters, y_filters, z_filters)
        self.conv_layers = []
        for i, (num_filters, kernel) in enumerate(zip(conv_layers, kernels)):
            self.conv_layers.append(
                conv_layer(
                    filters=num_filters, kernel_size=kernel, strides=conv_strides, use_bias=not conv_without_bias,
                    kernel_initializer=conv_kernel_initializer, bias_initializer=conv_bias_initializer,
                    padding='same', dilation_rate=2**i if conv_dilate else 1,
                ),
            )

        self.activations = [_activation_layer(activation) for _ in range(block_size)]
        self.normalizations = [_normalization_layer(conv_normalize) for _ in range(block_size)]
        self.regularizations = [_regularization_layer(self.tensor_map.axes(), conv_regularize, conv_regularize_rate) for _ in range(block_size)]
        self.fully_connected = DenseBlock(
            widths=dense_layers,
            activation=activation,
            normalization=dense_normalize,
            regularization=dense_regularize,
            regularization_rate=dense_regularize_rate,
            name=self.tensor_map.embed_name(),
        ) if dense_layers else None
        logging.info(f'Simple Block Convolutional Layers (num_filters, kernel_size): {list(zip(conv_layers, kernels))}')

    def can_apply(self):
        return self.tensor_map.axes() > 1

    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        if not self.can_apply():
            return x
        for convolve, activate, normalize, regularize in zip(
                self.conv_layers, self.activations, self.normalizations, self.regularizations,
        ):
            x = regularize(activate(normalize(convolve(x))))
        if self.fully_connected:
            x = Flatten()(x)
            x = self.fully_connected(x, intermediates)
        intermediates[self.tensor_map].append(x)
        return x


class ConvUp(Block):
    def __init__(
            self,
            *,
            tensor_map: TensorMap,
            conv_layers: List[int] = [32],
            conv_type: str = 'conv',
            conv_width: List[int] = [71],
            conv_x: List[int] = [3],
            conv_y: List[int] = [3],
            conv_z: List[int] = [3],
            pool_x: int = 2,
            pool_y: int = 2,
            pool_z: int = 1,
            activation: str = 'swish',
            conv_dilate: bool = False,
            conv_normalize: str = None,
            conv_regularize: str = None,
            conv_regularize_rate: float = 0.2,
            conv_without_bias: bool = False,
            conv_bias_initializer: str = 'zeros',
            conv_kernel_initializer: str = 'glorot_uniform',
            **kwargs,
    ):
        self.tensor_map = tensor_map
        if not self.can_apply():
            return
        x_filters, y_filters, z_filters = _get_xyz_filters(len(conv_layers), conv_x if self.tensor_map.axes() > 2 else conv_width, conv_y, conv_z)
        conv_layer, kernels = _conv_layer_from_kind_and_dimension(self.tensor_map.axes(), conv_type, x_filters, y_filters, z_filters)
        self.conv_layers = []
        for i, (num_filters, kernel) in enumerate(zip(conv_layers, kernels)):
            self.conv_layers.append(
                conv_layer(
                    filters=num_filters, kernel_size=kernel, use_bias=not conv_without_bias,
                    kernel_initializer=conv_kernel_initializer, bias_initializer=conv_bias_initializer,
                    padding='same', dilation_rate=2**i if conv_dilate else 1,
                ),
            )

        self.activations = [_activation_layer(activation) for _ in range(len(conv_layers))]
        self.normalizations = [_normalization_layer(conv_normalize) for _ in range(len(conv_layers))]
        self.upsamples = [_upsampler(tensor_map.axes(), pool_x, pool_y, pool_z) for _ in range(len(conv_layers))]
        self.regularizations = [_regularization_layer(self.tensor_map.axes(), conv_regularize, conv_regularize_rate) for _ in range(len(conv_layers))]
        self.start_shape = _start_shape_before_pooling(
            num_upsamples=len(conv_layers), output_shape=tensor_map.shape,
            upsample_rates=[pool_x, pool_y, pool_z], channels=conv_layers[-1],
        )
        self.reshape = FlatToStructure(output_shape=self.start_shape, activation=activation, normalization=conv_normalize)
        conv_layer, _ = _conv_layer_from_kind_and_dimension(tensor_map.axes(), 'conv', conv_x, conv_y, conv_z)
        self.conv_label = conv_layer(tensor_map.shape[-1], _one_by_n_kernel(tensor_map.axes()), activation=tensor_map.activation, name=tensor_map.output_name())
        logging.info(f'Simple Block Convolutional Layers (num_filters, kernel_size): {list(zip(conv_layers, kernels))}')

    def can_apply(self):
        return self.tensor_map.axes() > 1

    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        if not self.can_apply():
            return x
        if x.shape != self.start_shape:
            x = self.reshape(x)
        for convolve, activate, normalize, regularize, upsample in zip(
                self.conv_layers, self.activations, self.normalizations, self.regularizations, self.upsamples,
        ):
            x = regularize(activate(normalize(convolve(upsample(x)))))
        return self.conv_label(x)


class Residual:
    def __init__(
            self,
            *,
            dimension: int,
            filters_per_conv: List[int],
            conv_layer_type: str,
            conv_x: List[int],
            conv_y: List[int],
            conv_z: List[int],
            activation: str,
            normalization: str,
            regularization: str,
            regularization_rate: float,
            dilate: bool,
    ):
        block_size = len(filters_per_conv)
        assert len(conv_x) == len(conv_y) == len(conv_z) == block_size
        conv_layer, kernels = _conv_layer_from_kind_and_dimension(dimension, conv_layer_type, conv_x, conv_y, conv_z)
        self.conv_layers = []
        for i, (num_filters, kernel) in enumerate(zip(filters_per_conv, kernels)):
            if isinstance(conv_layer, DepthwiseConv2D):
                self.conv_layers.append(conv_layer(kernel_size=kernel, padding='same', dilation_rate=2 ** i if dilate else 1))
            else:
                self.conv_layers.append(conv_layer(filters=num_filters, kernel_size=kernel, padding='same', dilation_rate=2**i if dilate else 1))

        self.activations = [_activation_layer(activation) for _ in range(block_size)]
        self.normalizations = [_normalization_layer(normalization) for _ in range(block_size)]
        self.regularizations = [_regularization_layer(dimension, regularization, regularization_rate) for _ in range(block_size)]
        residual_conv_layer, _ = _conv_layer_from_kind_and_dimension(dimension, 'conv', conv_x, conv_y, conv_z)
        self.residual_convs = [residual_conv_layer(filters=filters_per_conv[0], kernel_size=_one_by_n_kernel(dimension)) for _ in range(block_size - 1)]
        logging.info(f'Residual Block Convolutional Layers (num_filters, kernel_size): {list(zip(filters_per_conv, kernels))}')

    def __call__(self, x: Tensor) -> Tensor:
        previous = x
        for convolve, activate, normalize, regularize, one_by_n_convolve in zip(
                self.conv_layers, self.activations, self.normalizations, self.regularizations, [None] + self.residual_convs,
        ):
            x = regularize(normalize(activate(convolve(x))))
            if one_by_n_convolve is not None:  # Do not residual add the input
                x = Add()([one_by_n_convolve(x), previous])
            previous = x
        return x


class DenseConvolutional:
    def __init__(
            self,
            *,
            dimension: int,
            block_size: int,
            conv_layer_type: str,
            filters: int,
            conv_x: List[int],
            conv_y: List[int],
            conv_z: List[int],
            activation: str,
            normalization: str,
            regularization: str,
            regularization_rate: float,
    ):
        conv_layer, kernels = _conv_layer_from_kind_and_dimension(dimension, conv_layer_type, conv_x, conv_y, conv_z)
        if isinstance(conv_layer, DepthwiseConv2D):
            self.conv_layers = [conv_layer(kernel_size=kernel, padding='same') for kernel in kernels]
        else:
            self.conv_layers = [conv_layer(filters=filters, kernel_size=kernel, padding='same') for kernel in kernels]
        self.activations = [_activation_layer(activation) for _ in range(block_size)]
        self.normalizations = [_normalization_layer(normalization) for _ in range(block_size)]
        self.regularizations = [_regularization_layer(dimension, regularization, regularization_rate) for _ in range(block_size)]
        logging.info(f'Dense Block Convolutional Layers (num_filters, kernel_size): {list(zip([filters]*len(kernels), kernels))}')

    def __call__(self, x: Tensor) -> Tensor:
        dense_connections = [x]
        for i, (convolve, activate, normalize, regularize) in enumerate(
            zip(
                    self.conv_layers, self.activations, self.normalizations, self.regularizations,
            ),
        ):
            x = normalize(regularize(activate(convolve(x))))
            if i < len(self.conv_layers) - 1:  # output of block does not get concatenated to
                dense_connections.append(x)
                x = Concatenate()(dense_connections[:])  # [:] is necessary because of tf weirdness
        return x


class FlatToStructure:
    """Takes a flat input, applies a dense layer, then restructures to output_shape"""
    def __init__(
            self,
            output_shape: Tuple[int, ...],
            activation: str,
            normalization: str,
    ):
        self.input_shapes = output_shape
        self.dense = Dense(units=int(np.prod(output_shape)))
        self.activation = _activation_layer(activation)
        self.reshape = Reshape(output_shape)
        self.norm = _normalization_layer(normalization)

    def __call__(self, x: Tensor) -> Tensor:
        return self.reshape(self.norm(self.activation(self.dense(x))))


def _repeat_dimension(filters: List[int], num_filters_needed: int) -> List[int]:
    if len(filters) < num_filters_needed:
        repeat = num_filters_needed // len(filters) + 1
        filters = (filters * repeat)[:num_filters_needed]
    return filters


def _start_shape_before_pooling(
        num_upsamples: int, output_shape: Tuple[int, ...], upsample_rates: Sequence[int], channels: int,
) -> Tuple[int, ...]:
    """
    Given the number of blocks in the decoder and the upsample rates, return required input shape to get to output shape
    """
    upsample_rates = list(upsample_rates) + [1] * len(output_shape)
    return tuple((shape // rate**num_upsamples for shape, rate in zip(output_shape[:-1], upsample_rates))) + (channels,)


def _get_xyz_filters(num_filters, conv_x, conv_y, conv_z):
    # list of filter dimensions should match the total number of convolutional layers
    x_filters = _repeat_dimension(conv_x, num_filters)
    y_filters = _repeat_dimension(conv_y, num_filters)
    z_filters = _repeat_dimension(conv_z, num_filters)
    return x_filters, y_filters, z_filters



# tensor_writer_ukbb.py
#
# UK Biobank-specific tensor writing, SQL querying, data munging goes here
#

# Imports
import os
import re
import csv
import glob
import time
import shutil
import logging
import datetime
import operator
import tempfile
import traceback
from functools import partial
from itertools import product
from collections import Counter, defaultdict
from typing import Dict, List, Tuple, Optional, Any

import h5py
import imageio
import pydicom
import sqlite3
import zipfile
import matplotlib
matplotlib.use('Agg')
import numpy as np
import nibabel as nib
import matplotlib.pyplot as plt
from PIL import Image, ImageDraw  # Polygon to mask
import xml.etree.ElementTree as et
from timeit import default_timer as timer
from scipy.ndimage.morphology import binary_closing, binary_erosion  # Morphological operator

from ml4h.plots import plot_value_counter, plot_histograms
from ml4h.defines import ECG_BIKE_LEADS, ECG_BIKE_MEDIAN_SIZE, ECG_BIKE_STRIP_SIZE, ECG_BIKE_FULL_SIZE, MRI_FRAMES
from ml4h.defines import MRI_TO_SEGMENT, MRI_SEGMENTED_CHANNEL_MAP, MRI_ANNOTATION_CHANNEL_MAP, MRI_ANNOTATION_NAME
from ml4h.defines import StorageType, IMAGE_EXT, TENSOR_EXT, DICOM_EXT, JOIN_CHAR, CONCAT_CHAR, HD5_GROUP_CHAR, DATE_FORMAT
from ml4h.defines import MRI_PIXEL_WIDTH, MRI_PIXEL_HEIGHT, MRI_SLICE_THICKNESS, MRI_PATIENT_ORIENTATION, MRI_PATIENT_POSITION



MRI_MIN_RADIUS = 2
MRI_MAX_MYOCARDIUM = 20
MRI_BIG_RADIUS_FACTOR = 0.9
MRI_SMALL_RADIUS_FACTOR = 0.19
MRI_MITRAL_VALVE_THICKNESS = 6
MRI_SWI_SLICES_TO_AXIS_SHIFT = 48
MRI_CARDIAC_SERIES = [
    'cine_segmented_lax_2ch', 'cine_segmented_lax_3ch', 'cine_segmented_lax_4ch', 'cine_segmented_sax_b1', 'cine_segmented_sax_b2',
    'cine_segmented_sax_b3', 'cine_segmented_sax_b4', 'cine_segmented_sax_b5', 'cine_segmented_sax_b6', 'cine_segmented_sax_b7',
    'cine_segmented_sax_b8', 'cine_segmented_sax_b9', 'cine_segmented_sax_b10', 'cine_segmented_sax_b11', 'cine_segmented_sax_b12',
    'cine_segmented_sax_b13', 'cine_segmented_sax_inlinevf', 'cine_segmented_lax_inlinevf', 'cine_segmented_ao_dist',
    'cine_segmented_lvot', 'flow_250_tp_aov_bh_epat@c_p', 'flow_250_tp_aov_bh_epat@c', 'flow_250_tp_aov_bh_epat@c_mag',
    'shmolli_192i_b1_sax_b1s_sax_b1s_sax_b1s_t1map', 'shmolli_192i_sax_b2s_sax_b2s_sax_b2s_t1map',
    'shmolli_192i_b2_sax_b2s_sax_b2s_sax_b2s_t1map',
    'shmolli_192i_b3_sax_b3s_sax_b3s_sax_b3s_t1map', 'shmolli_192i_b4_sax_b4s_sax_b4s_sax_b4s_t1map',
    'shmolli_192i_b5_sax_b5s_sax_b5s_sax_b5s_t1map', 'shmolli_192i_b6_sax_b6s_sax_b6s_sax_b6s_t1map',
    'shmolli_192i_b7_sax_b7s_sax_b7s_sax_b7s_t1map',

]
MRI_CARDIAC_SERIES_SEGMENTED = [series+'_segmented' for series in MRI_CARDIAC_SERIES]
MRI_BRAIN_SERIES = ['t1_p2_1mm_fov256_sag_ti_880', 't2_flair_sag_p2_1mm_fs_ellip_pf78']
MRI_NIFTI_FIELD_ID_TO_ROOT = {'20251': 'SWI', '20252': 'T1', '20253': 'T2_FLAIR'}
MRI_LIVER_SERIES = ['gre_mullti_echo_10_te_liver', 'lms_ideal_optimised_low_flip_6dyn', 'shmolli_192i', 'shmolli_192i_liver', 'shmolli_192i_fitparams', 'shmolli_192i_t1map']
MRI_LIVER_SERIES_12BIT = ['gre_mullti_echo_10_te_liver_12bit', 'lms_ideal_optimised_low_flip_6dyn_12bit', 'shmolli_192i_12bit', 'shmolli_192i_liver_12bit']
MRI_LIVER_IDEAL_PROTOCOL = ['lms_ideal_optimised_low_flip_6dyn', 'lms_ideal_optimised_low_flip_6dyn_12bit']

DICOM_MRI_FIELDS = [
    '20209', '20208', '20210', '20212', '20213', '20214', '20204', '20203', '20254', '20216', '20220', '20218',
    '20227', '20225', '20217', '20158',
]

DXA_FIELD = '20158'

ECG_BIKE_FIELD = '6025'
ECG_REST_FIELD = '20205'
ECG_TABLE_TAGS = ['RAmplitude', 'SAmplitude']
ECG_TAGS_TO_WRITE = [
    'VentricularRate', 'PQInterval', 'PDuration', 'QRSDuration', 'QTInterval', 'QTCInterval', 'RRInterval', 'PPInterval',
    'SokolovLVHIndex', 'PAxis', 'RAxis', 'TAxis', 'QTDispersion', 'QTDispersionBazett', 'QRSNum', 'POnset', 'POffset', 'QOnset',
    'QOffset', 'TOffset',
]
ECG_BIKE_SAMPLE_RATE = 500
ECG_BIKE_NUM_LEADS = 3
SECONDS_PER_MINUTE = 60


def write_tensors(
    a_id: str,
    xml_folder: str,
    zip_folder: str,
    output_folder: str,
    tensors: str,
    mri_unzip: str,
    mri_field_ids: List[int],
    xml_field_ids: List[int],
    write_pngs: bool,
    min_sample_id: int,
    max_sample_id: int,
    min_values_to_print: int,
) -> None:
    """Write tensors as HD5 files containing any kind of data from UK BioBank

    One HD5 file is generated per sample.  Each file may contain many tensor encodings of data including:
     survey responses, MRI, and ECG.

    :param a_id: User chosen string to identify this run
    :param xml_folder: Path to folder containing ECG XML files
    :param zip_folder: Path to folder containing zipped DICOM files
    :param output_folder: Folder to write outputs to (mostly for debugging)
    :param tensors: Folder to populate with HD5 tensors
    :param mri_unzip: Folder where zipped DICOM will be decompressed
    :param mri_field_ids: List of MRI field IDs from UKBB
    :param xml_field_ids: List of ECG field IDs from UKBB
    :param write_pngs: write MRIs as PNG images for debugging
    :param min_sample_id: Minimum sample id to generate, for parallelization
    :param max_sample_id: Maximum sample id to generate, for parallelization
    :param min_values_to_print: Minimum number of samples that have responded to question for it to be included in the
            categorical or continuous dictionaries printed after tensor generation

    :return: None
    """
    stats = Counter()
    continuous_stats = defaultdict(list)
    sample_ids = range(min_sample_id, max_sample_id)
    for sample_id in sorted(sample_ids):

        start_time = timer()  # Keep track of elapsed execution time
        tp = os.path.join(tensors, str(sample_id) + TENSOR_EXT)
        if not os.path.exists(os.path.dirname(tp)):
            os.makedirs(os.path.dirname(tp))
        if _prune_sample(sample_id, min_sample_id, max_sample_id, mri_field_ids, xml_field_ids, zip_folder, xml_folder):
            continue
        try:
            with h5py.File(tp, 'w') as hd5:
                _write_tensors_from_zipped_dicoms(write_pngs, tensors, mri_unzip, mri_field_ids, zip_folder, hd5, sample_id, stats)
                _write_tensors_from_zipped_niftis(zip_folder, mri_field_ids, hd5, sample_id, stats)
                _write_tensors_from_xml(xml_field_ids, xml_folder, hd5, sample_id, write_pngs, stats, continuous_stats)
                stats['Tensors written'] += 1
        except AttributeError:
            logging.exception('Encountered AttributeError trying to write a UKBB tensor at path:{}'.format(tp))
            logging.info('Deleting attempted tensor at path:{}'.format(tp))
            os.remove(tp)
        except ValueError:
            logging.exception('Encountered ValueError trying to write a UKBB tensor at path:{}'.format(tp))
            logging.info('Deleting attempted tensor at path:{}'.format(tp))
            os.remove(tp)
        except RuntimeError:
            logging.exception('Encountered RuntimeError trying to write a UKBB tensor at path:{}'.format(tp))
            logging.info('Deleting attempted tensor at path:{}'.format(tp))
            os.remove(tp)
        except IndexError:
            logging.exception('Encountered IndexError trying to write a UKBB tensor at path:{}'.format(tp))
            logging.info('Deleting attempted tensor at path:{}'.format(tp))
            os.remove(tp)
        except OSError:
            logging.exception('Encountered OSError trying to write a UKBB tensor at path:{}'.format(tp))
            logging.info('Deleting attempted tensor at path:{}'.format(tp))
            os.remove(tp)

        end_time = timer()
        elapsed_time = end_time - start_time
        logging.info(f"Populated {tp} in {elapsed_time:0.2f} seconds.")

    _dicts_and_plots_from_tensorization(a_id, output_folder, min_values_to_print, write_pngs, continuous_stats, stats)


def write_tensors_from_dicom_pngs(
    tensors, png_path, manifest_tsv, series, min_sample_id, max_sample_id, x=256, y=256,
    sample_header='sample_id', dicom_header='dicom_file',
    instance_header='instance_number', png_postfix='.png.mask.png',
    path_prefix='ukb_cardiac_mri',
):
    stats = Counter()
    reader = csv.reader(open(manifest_tsv), delimiter='\t')
    header = next(reader)
    logging.info(f"DICOM Manifest Header is:{header}")
    instance_index = header.index(instance_header)
    sample_index = header.index(sample_header)
    dicom_index = header.index(dicom_header)
    for row in reader:
        sample_id = row[sample_index]
        if not min_sample_id <= int(sample_id) < max_sample_id:
            continue
        stats[sample_header + '_' + sample_id] += 1
        dicom_file = row[dicom_index]
        try:
            png = imageio.imread(os.path.join(png_path, dicom_file + png_postfix))
            full_tensor = np.zeros((x, y), dtype=np.float32)
            full_tensor[:png.shape[0], :png.shape[1]] = png[..., 0]
            tensor_file = os.path.join(tensors, str(sample_id) + TENSOR_EXT)
            if not os.path.exists(os.path.dirname(tensor_file)):
                os.makedirs(os.path.dirname(tensor_file))
            with h5py.File(tensor_file, 'a') as hd5:
                tensor_name = series + '_annotated_' + row[instance_index]
                tp = tensor_path(path_prefix, tensor_name)
                if tp in hd5:
                    tensor = first_dataset_at_path(hd5, tp)
                    tensor[:] = full_tensor
                    stats['updated'] += 1
                else:
                    create_tensor_in_hd5(hd5, path_prefix, tensor_name, full_tensor, stats)
                    stats['created'] += 1

        except FileNotFoundError:
            logging.warning(f'Could not find file: {os.path.join(png_path, dicom_file + png_postfix)}')
            stats['File not found error'] += 1
    for k in stats:
        if sample_header in k and stats[k] == 50:
            continue
        logging.info(f'{k} has {stats[k]}')


def write_tensors_from_ecg_pngs(tensors, png_path, min_sample_id, max_sample_id, name='ecg_segmented', png_postfix='.png.mask.png', path_prefix='ecg_rest'):
    stats = Counter()
    for png_file in os.listdir(png_path):
        sample_id = png_file.split('.')[0]
        offset_seconds = float(png_file.split('_')[-1].replace(png_postfix, ''))
        if not min_sample_id <= int(sample_id) < max_sample_id:
            continue

        png = imageio.imread(os.path.join(png_path, png_file))
        stats[f'png shape {png.shape}'] += 1
        ecg_1d_segmentation = png[0, :, 0]
        logging.info(f' ecg1d segmentation unique: {np.unique(ecg_1d_segmentation)} \nstart {ecg_1d_segmentation[:60]} ')
        stats[f'ecg_1d_segmentation shape {ecg_1d_segmentation.shape}'] += 1
        tensor_file = os.path.join(tensors, str(sample_id) + TENSOR_EXT)
        if not os.path.exists(os.path.dirname(tensor_file)):
            os.makedirs(os.path.dirname(tensor_file))
        with h5py.File(tensor_file, 'a') as hd5:
            tp = tensor_path(path_prefix, name)
            if tp in hd5:
                tensor_dataset = first_dataset_at_path(hd5, tp)
                tensor_dataset[:] = ecg_1d_segmentation
                tensor_dataset.attrs['offset_seconds'] = offset_seconds
                stats['updated'] += 1
            else:
                create_tensor_in_hd5(hd5, path_prefix, name, ecg_1d_segmentation, stats, attributes={'offset_seconds': offset_seconds})
                stats['created'] += 1

    for k in stats:
        logging.info(f'{k} has {stats[k]}')


def _load_meta_data_for_tensor_writing(volume_csv: str, lv_mass_csv: str, min_sample_id: int, max_sample_id: int) -> Tuple[Dict[int, Dict[str, float]], List[int]]:
    """ Gather metadata necessary to write tensors from UK biobank

    Loads the field IDs of survey data, dates of assessment, diagnosis of diseases,
    ejection fractions, diastolic volumes, systolic volumes, and sample IDs to make tensors from.

    :param volume_csv: CSV containing systole and diastole volumes and ejection fraction for samples with MRI
    :param lv_mass_csv: TSV containing left ventricular mass and other cardiac MRI readouts on ~5000 people returned from app 2964
    :param min_sample_id: Minimum sample id to generate, for parallelization
    :param max_sample_id: Maximum sample id to generate, for parallelization
    :return: Tuple of metadata containers
        nested_dictionary: Dictionary mapping sample IDs (as ints) to dictionaries mapping strings to values from the CSV
        sample_ids: List of sample IDs (as ints) to generate tensors for
    """
    nested_dictionary = defaultdict(dict)
    with open(volume_csv, 'r') as volumes:
        lol = list(csv.reader(volumes, delimiter='\t'))
        logging.info(f"CSV of MRI volumes header:{list(enumerate(lol[0]))}")
        fields = lol[0][1:]  # Assumes sample id is the first field
        for row in lol[1:]:
            sample_id = int(row[0])
            if min_sample_id <= sample_id <= max_sample_id:
                nested_dictionary[sample_id] = {fields[i].strip().lower(): row[i+1] for i in range(len(fields))}

    with open(lv_mass_csv, 'r') as lvm:
        lol = list(csv.reader(lvm, delimiter='\t'))
        logging.info('CSV of returned MRI mass, etc, header:{}'.format(list(enumerate(lol[0]))))
        for row in lol[1:]:
            # column 0 is the original app's sample ID. Column 1 is app 7089's sample ID.
            sample_id = int(row[1])
            if min_sample_id <= sample_id <= max_sample_id and len(row[13]) > 0 and row[13] != 'NA':
                # Zero-based column #13 is the LV mass
                nested_dictionary[sample_id]['lv_mass'] = float(row[13])

    return nested_dictionary


def _sample_has_dicom_mris(zip_folder, sample_id) -> bool:
    sample_str = str(sample_id)
    return any([os.path.exists(os.path.join(zip_folder, f'{sample_str}_{mri_f}_2_0.zip')) for mri_f in DICOM_MRI_FIELDS])


def _sample_has_nifti_mris(zip_folder, sample_id) -> bool:
    sample_str = str(sample_id)
    return any([os.path.exists(os.path.join(zip_folder, f'{sample_str}_{mri_f}_2_0.zip')) for mri_f in MRI_NIFTI_FIELD_ID_TO_ROOT])


def _sample_has_ecgs(xml_folder, xml_field_ids, sample_id) -> bool:
    sample_str = str(sample_id)
    for xml_id in xml_field_ids:
        if os.path.exists(xml_folder + sample_str + '_' + xml_id + '_0_0.xml'):
            return True
        if os.path.exists(xml_folder + sample_str + '_' + xml_id + '_1_0.xml'):
            return True
        if os.path.exists(xml_folder + sample_str + '_' + xml_id + '_2_0.xml'):
            return True
    return False


def _dicts_and_plots_from_tensorization(
    a_id: str,
    output_folder: str,
    min_values_to_print: int,
    write_pngs: bool,
    continuous_stats: Dict[str, List[float]],
    stats: Dict[str, int],
) -> None:
    """Print out dictionaries of data encountered during tensorization. Optionally make plots of this data.

    :param a_id: User chosen string to identify this run
    :param output_folder: Folder to write outputs to (mostly for debugging)
    :param min_values_to_print: Minimum number of samples that have responded to question for it to be included in the
            categorical or continuous dictionaries printed after tensor generation
    :param write_pngs: write MRIs as PNG images for debugging
    :param continuous_stats: Dictionary mapping field meanings to the list of continuous values found for them
    :param stats: Dictionary mapping strings to ints keeps track of categorical responses and other info.
    :return: None
    """
    categories = {}
    continuous = {}
    value_counter = Counter()
    for k in sorted(list(stats.keys())):
        logging.info("{} has {}".format(k, stats[k]))

        if 'categorical' not in k and 'continuous' not in k:
            continue
        parts = k.split(JOIN_CHAR)
        if len(parts) > 1:
            column_key = k.replace(HD5_GROUP_CHAR, JOIN_CHAR)  # flatten the hd5 group
            value_counter[column_key] += stats[k]
            if min_values_to_print < value_counter[column_key]:
                if 'categorical' in k and column_key not in categories:
                    categories[column_key.replace('categorical_', '')] = len(categories)
                if 'continuous' in k and column_key not in continuous:
                    continuous[column_key.replace('continuous_', '')] = len(continuous)

    if write_pngs:
        plot_value_counter(list(categories.keys()), value_counter, a_id + '_v_count', os.path.join(output_folder, a_id))
        plot_histograms(continuous_stats, a_id, os.path.join(output_folder, a_id))

    logging.info("Continuous tensor map: {}".format(continuous))
    logging.info("Continuous Columns: {}".format(len(continuous)))
    logging.info("Category tensor map: {}".format(categories))
    logging.info("Categories Columns: {}".format(len(categories)))


def _to_float_or_false(s):
    try:
        return float(s)
    except ValueError:
        return False


def _to_float_or_nan(s):
    try:
        return float(s)
    except ValueError:
        return np.nan


def _write_tensors_from_zipped_dicoms(
    write_pngs: bool,
    tensors: str,
    dicoms: str,
    mri_field_ids: List[str],
    zip_folder: str,
    hd5: h5py.File,
    sample_id: int,
    stats: Dict[str, int],
) -> None:
    sample_str = str(sample_id)
    for mri_field in set(mri_field_ids).intersection(DICOM_MRI_FIELDS):
        mris = glob.glob(zip_folder + sample_str + '_' + mri_field + '*.zip')
        for zipped in sorted(mris):
            instance = zipped.split(JOIN_CHAR)[-2] # UKB Bulk filename: sampleid_fieldid_instance_arrayidx
            logging.info(f"Got zipped dicoms for sample: {sample_id}  MRI field: {mri_field} from instance {instance}")
            dicom_folder = os.path.join(dicoms, sample_str, mri_field)
            if not os.path.exists(dicom_folder):
                os.makedirs(dicom_folder)
            with zipfile.ZipFile(zipped, "r") as zip_ref:
                zip_ref.extractall(dicom_folder)
                _write_tensors_from_dicoms(
                    write_pngs, tensors, dicom_folder,
                    hd5, instance, stats,
                )
                stats['MRI fields written'] += 1
            shutil.rmtree(dicom_folder)


def _write_tensors_from_zipped_niftis(zip_folder: str, mri_field_ids: List[str], hd5: h5py.File, sample_id: str, stats: Dict[str, int]) -> None:
    for mri_field in set(mri_field_ids).intersection(set(MRI_NIFTI_FIELD_ID_TO_ROOT.keys())):
        mris = glob.glob(os.path.join(zip_folder, f'{sample_id}_{mri_field}*.zip'))
        for zipped in mris:
            logging.info(f"Got zipped niftis for sample: {sample_id} with MRI field: {mri_field}")
            with tempfile.TemporaryDirectory() as temp_folder, zipfile.ZipFile(zipped, "r") as zip_ref:
                zip_ref.extractall(temp_folder)
                _write_tensors_from_niftis(temp_folder, hd5, mri_field, stats)
                stats['MRI fields written'] += 1


def _write_tensors_from_dicoms(
    write_pngs: bool, tensors: str,
    dicom_folder: str, hd5: h5py.File, instance: str, stats: Dict[str, int],
) -> None:
    """Convert a folder of DICOMs from a sample into tensors for each series

    Segmented dicoms require special processing and are written to tensor per-slice

    Arguments
        :param write_pngs: write MRIs as PNG images for debugging
        :param tensors: Folder where hd5 tensor files are being written
        :param dicom_folder: Folder with all dicoms associated with one sample.
        :param hd5: Tensor file in which to create datasets for each series and each segmented slice
        :param instance: The current instance index as a string
        :param stats: Counter to keep track of summary statistics

    """
    views = defaultdict(list)
    min_ideal_series = 9e9
    for dicom in os.listdir(dicom_folder):
        if os.path.splitext(dicom)[-1] != DICOM_EXT:
            continue
        d = pydicom.read_file(os.path.join(dicom_folder, dicom))
        series = d.SeriesDescription.lower().replace(' ', '_')
        if series + '_12bit' in MRI_LIVER_SERIES_12BIT and d.LargestImagePixelValue > 2048:
            views[series + '_12bit'].append(d)
            stats[series + '_12bit'] += 1
        elif series in MRI_LIVER_SERIES + MRI_CARDIAC_SERIES + MRI_BRAIN_SERIES:
            views[series].append(d)
            stats[series] += 1
        elif series == 'dxa_images':
            series_num = dicom.split('.')[-5]
            dxa_number = dicom.split('.')[-4]
            name = f'dxa_{series_num}_{dxa_number}'
            create_tensor_in_hd5(hd5, f'ukb_dxa/', name, d.pixel_array, stats)

        if series in MRI_LIVER_IDEAL_PROTOCOL:
            min_ideal_series = min(min_ideal_series, int(d.SeriesNumber))

    for v in views:
        mri_shape = (views[v][0].Rows, views[v][0].Columns, len(views[v]))
        mri_date = _datetime_from_dicom(views[v][0])
        stats[v + ' mri shape:' + str(mri_shape)] += 1
        if v in MRI_BRAIN_SERIES:
            mri_group = 'ukb_brain_mri'
        elif v in MRI_LIVER_SERIES + MRI_LIVER_SERIES_12BIT:
            mri_group = 'ukb_liver_mri'
        elif v in MRI_CARDIAC_SERIES + MRI_CARDIAC_SERIES_SEGMENTED:
            mri_group = 'ukb_cardiac_mri'
        else:
            mri_group = 'ukb_mri'

        if v == MRI_TO_SEGMENT:
            _tensorize_short_and_long_axis_segmented_cardiac_mri(views[v], v, write_pngs, tensors, hd5, mri_date, mri_group, stats)
        elif v in MRI_BRAIN_SERIES:
            _tensorize_brain_mri(views[v], v, mri_date, mri_group, hd5)

        else:
            mri_data = np.zeros((views[v][0].Rows, views[v][0].Columns, len(views[v])), dtype=np.float32)
            for slicer in views[v]:
                _save_pixel_dimensions_if_missing(slicer, v, hd5)
                _save_slice_thickness_if_missing(slicer, v, hd5)
                _save_series_orientation_and_position_if_missing(slicer, v, hd5)
                slice_index = slicer.InstanceNumber - 1
                if v in MRI_LIVER_IDEAL_PROTOCOL:
                    slice_index = _slice_index_from_ideal_protocol(slicer, min_ideal_series)
                mri_data[..., slice_index] = slicer.pixel_array.astype(np.float32)
            create_tensor_in_hd5(hd5, mri_group, v, mri_data, stats, mri_date, instance=instance)


def _tensorize_short_and_long_axis_segmented_cardiac_mri(
    slices: List[pydicom.Dataset], series: str,
    write_pngs: bool, tensors: str,
    hd5: h5py.File, mri_date: datetime.datetime, mri_group: str,
    stats: Dict[str, int],
) -> None:
    systoles = {}
    diastoles = {}
    systoles_pix = {}
    systoles_masks = {}
    diastoles_masks = {}

    for slicer in slices:
        full_mask = np.zeros((slicer.Rows, slicer.Columns), dtype=np.float32)
        full_slice = np.zeros((slicer.Rows, slicer.Columns), dtype=np.float32)

        if _has_overlay(slicer):
            if _is_mitral_valve_segmentation(slicer):
                series = series.replace('sax', 'lax')
            else:
                series = series.replace('lax', 'sax')
            series_segmented = f'{series}_segmented'

            try:
                overlay, mask, ventricle_pixels, _ = _get_overlay_from_dicom(slicer)
            except KeyError:
                logging.exception(f'Got key error trying to make anatomical mask, skipping.')
                continue

            _save_pixel_dimensions_if_missing(slicer, series, hd5)
            _save_slice_thickness_if_missing(slicer, series, hd5)
            _save_series_orientation_and_position_if_missing(slicer, series, hd5, str(slicer.InstanceNumber))
            _save_pixel_dimensions_if_missing(slicer, series_segmented, hd5)
            _save_slice_thickness_if_missing(slicer, series_segmented, hd5)
            _save_series_orientation_and_position_if_missing(slicer, series_segmented, hd5, str(slicer.InstanceNumber))

            cur_angle = (slicer.InstanceNumber - 1) // MRI_FRAMES  # dicom InstanceNumber is 1-based
            full_slice[:] = slicer.pixel_array.astype(np.float32)
            create_tensor_in_hd5(hd5, mri_group, f'{series}{HD5_GROUP_CHAR}{slicer.InstanceNumber}', full_slice, stats, mri_date)

            if (slicer.InstanceNumber - 1) % MRI_FRAMES == 0:  # Diastole frame is always the first
                diastoles[cur_angle] = slicer
                diastoles_masks[cur_angle] = mask
            if cur_angle not in systoles:
                systoles[cur_angle] = slicer
                systoles_pix[cur_angle] = ventricle_pixels
                systoles_masks[cur_angle] = mask
            else:
                if ventricle_pixels < systoles_pix[cur_angle]:
                    systoles[cur_angle] = slicer
                    systoles_pix[cur_angle] = ventricle_pixels
                    systoles_masks[cur_angle] = mask

    for angle in diastoles:
        logging.info(f'Found systole, instance:{systoles[angle].InstanceNumber} ventricle pixels:{systoles_pix[angle]}')
        full_slice = diastoles[angle].pixel_array.astype(np.float32)
        create_tensor_in_hd5(hd5, mri_group, f'diastole_frame_b{angle}', full_slice, stats, mri_date)
        create_tensor_in_hd5(hd5, mri_group, f'diastole_mask_b{angle}', diastoles_masks[angle], stats, mri_date)
        if write_pngs:
            plt.imsave(tensors + 'diastole_frame_b' + str(angle) + IMAGE_EXT, full_slice)
            plt.imsave(tensors + 'diastole_mask_b' + str(angle) + IMAGE_EXT, full_mask)

        full_slice = systoles[angle].pixel_array.astype(np.float32)
        create_tensor_in_hd5(hd5, mri_group, f'systole_frame_b{angle}', full_slice, stats, mri_date)
        create_tensor_in_hd5(hd5, mri_group, f'systole_mask_b{angle}', systoles_masks[angle], stats, mri_date)
        if write_pngs:
            plt.imsave(tensors + 'systole_frame_b' + str(angle) + IMAGE_EXT, full_slice)
            plt.imsave(tensors + 'systole_mask_b' + str(angle) + IMAGE_EXT, full_mask)


def _tensorize_brain_mri(slices: List[pydicom.Dataset], series: str, mri_date: datetime.datetime, mri_group: str, hd5: h5py.File) -> None:
    mri_data1 = np.zeros((slices[0].Rows, slices[0].Columns, len(slices) // 2), dtype=np.float32)
    mri_data2 = np.zeros((slices[0].Rows, slices[0].Columns, len(slices) // 2), dtype=np.float32)
    for slicer in slices:
        _save_pixel_dimensions_if_missing(slicer, series, hd5)
        _save_slice_thickness_if_missing(slicer, series, hd5)
        _save_series_orientation_and_position_if_missing(slicer, series, hd5)
        slice_index = slicer.InstanceNumber - 1
        if slicer.SeriesNumber % 2 == 0:  # Assumes one series is even and one series number is odd
            mri_data1[..., slice_index] = slicer.pixel_array.astype(np.float32)
        else:
            mri_data2[..., slice_index] = slicer.pixel_array.astype(np.float32)

    create_tensor_in_hd5(hd5, mri_group, series + '_1', mri_data1, date=mri_date)
    create_tensor_in_hd5(hd5, mri_group, series + '_2', mri_data2, date=mri_date)


def _save_pixel_dimensions_if_missing(slicer, series, hd5):
    if MRI_PIXEL_WIDTH + '_' + series not in hd5 and series in MRI_BRAIN_SERIES + MRI_CARDIAC_SERIES + MRI_CARDIAC_SERIES_SEGMENTED + MRI_LIVER_SERIES + MRI_LIVER_SERIES_12BIT:
        hd5.create_dataset(MRI_PIXEL_WIDTH + '_' + series, data=float(slicer.PixelSpacing[0]))
    if MRI_PIXEL_HEIGHT + '_' + series not in hd5 and series in MRI_BRAIN_SERIES + MRI_CARDIAC_SERIES + MRI_CARDIAC_SERIES_SEGMENTED + MRI_LIVER_SERIES + MRI_LIVER_SERIES_12BIT:
        hd5.create_dataset(MRI_PIXEL_HEIGHT + '_' + series, data=float(slicer.PixelSpacing[1]))


def _save_slice_thickness_if_missing(slicer, series, hd5):
    if MRI_SLICE_THICKNESS + '_' + series not in hd5 and series in MRI_BRAIN_SERIES + MRI_CARDIAC_SERIES + MRI_CARDIAC_SERIES_SEGMENTED + MRI_LIVER_SERIES + MRI_LIVER_SERIES_12BIT:
        hd5.create_dataset(MRI_SLICE_THICKNESS + '_' + series, data=float(slicer.SliceThickness))


def _save_series_orientation_and_position_if_missing(slicer, series, hd5, instance=None):
    orientation_ds_name = MRI_PATIENT_ORIENTATION + '_' + series
    position_ds_name = MRI_PATIENT_POSITION + '_' + series
    if instance:
        orientation_ds_name += HD5_GROUP_CHAR + instance
        position_ds_name += HD5_GROUP_CHAR + instance
    if orientation_ds_name not in hd5 and series in MRI_BRAIN_SERIES + MRI_CARDIAC_SERIES + MRI_CARDIAC_SERIES_SEGMENTED + MRI_LIVER_SERIES + MRI_LIVER_SERIES_12BIT:
        hd5.create_dataset(orientation_ds_name, data=[float(x) for x in slicer.ImageOrientationPatient])
    if position_ds_name not in hd5 and series in MRI_BRAIN_SERIES + MRI_CARDIAC_SERIES + MRI_CARDIAC_SERIES_SEGMENTED + MRI_LIVER_SERIES + MRI_LIVER_SERIES_12BIT:
        hd5.create_dataset(position_ds_name, data=[float(x) for x in slicer.ImagePositionPatient])


def _has_overlay(d) -> bool:
    try:
        _ = d[0x6000, 0x3000].value
        return True
    except KeyError:
        return False


def _is_mitral_valve_segmentation(d) -> bool:
    return d.SliceThickness == 6


def _slice_index_from_ideal_protocol(d, min_ideal_series):
    return 6*(d.InstanceNumber-1) + ((d.SeriesNumber-min_ideal_series)//2)


def _get_overlay_from_dicom(d, debug=False) -> Tuple[np.ndarray, np.ndarray]:
    """Get an overlay from a DICOM file

    Morphological operators are used to transform the pixel outline of the myocardium
    to the labeled pixel masks for myocardium and left ventricle

    Arguments
        d: the dicom file
        stats: Counter to keep track of summary statistics

    Returns
        Tuple of two numpy arrays.
        The first is the raw overlay array with myocardium outline,
        The second is a pixel mask with 0 for background 1 for myocardium and 2 for ventricle
    """
    i_overlay = 0
    dicom_tag = 0x6000 + 2 * i_overlay
    overlay_raw = d[dicom_tag, 0x3000].value
    rows = d[dicom_tag, 0x0010].value  # rows = 512
    cols = d[dicom_tag, 0x0011].value  # cols = 512
    overlay_frames = d[dicom_tag, 0x0015].value
    bits_allocated = d[dicom_tag, 0x0100].value

    np_dtype = np.dtype('uint8')
    length_of_pixel_array = len(overlay_raw)
    expected_length = rows * cols
    if bits_allocated == 1:
        expected_bit_length = expected_length
        bit = 0
        overlay = np.ndarray(shape=(length_of_pixel_array * 8), dtype=np_dtype)
        for byte in overlay_raw:
            for bit in range(bit, bit + 8):
                overlay[bit] = byte & 0b1
                byte >>= 1
            bit += 1
        overlay = overlay[:expected_bit_length]
    if overlay_frames == 1:
        overlay = overlay.reshape(rows, cols)
        idx = np.where(overlay == 1)
        min_pos = (np.min(idx[0]), np.min(idx[1]))
        max_pos = (np.max(idx[0]), np.max(idx[1]))
        short_side = min((max_pos[0] - min_pos[0]), (max_pos[1] - min_pos[1]))
        small_radius = max(MRI_MIN_RADIUS, short_side * MRI_SMALL_RADIUS_FACTOR)
        big_radius = max(MRI_MIN_RADIUS+1, short_side * MRI_BIG_RADIUS_FACTOR)
        small_structure = _unit_disk(small_radius)
        m1 = binary_closing(overlay, small_structure).astype(np.int)
        big_structure = _unit_disk(big_radius)
        m2 = binary_closing(overlay, big_structure).astype(np.int)
        anatomical_mask = m1 + m2
        ventricle_pixels = np.count_nonzero(anatomical_mask == MRI_SEGMENTED_CHANNEL_MAP['ventricle'])
        myocardium_pixels = np.count_nonzero(anatomical_mask == MRI_SEGMENTED_CHANNEL_MAP['myocardium'])
        if ventricle_pixels == 0 and myocardium_pixels > MRI_MAX_MYOCARDIUM:  # try to rescue small ventricles
            erode_structure = _unit_disk(small_radius*1.5)
            anatomical_mask = anatomical_mask - binary_erosion(m1, erode_structure).astype(np.int)
            ventricle_pixels = np.count_nonzero(anatomical_mask == MRI_SEGMENTED_CHANNEL_MAP['ventricle'])
            myocardium_pixels = np.count_nonzero(anatomical_mask == MRI_SEGMENTED_CHANNEL_MAP['myocardium'])
        return overlay, anatomical_mask, ventricle_pixels, myocardium_pixels


def _unit_disk(r) -> np.ndarray:
    y, x = np.ogrid[-r: r + 1, -r: r + 1]
    return (x ** 2 + y ** 2 <= r ** 2).astype(np.int)


def _outline_to_mask(labeled_outline, idx) -> np.ndarray:
    idx = np.where(labeled_outline == idx)
    poly = list(zip(idx[1].tolist(), idx[0].tolist()))
    img = Image.new("L", [labeled_outline.shape[1], labeled_outline.shape[0]], 0)
    ImageDraw.Draw(img).polygon(poly, outline=1, fill=1)
    return np.array(img)


def _write_tensors_from_xml(xml_field_ids, xml_folder, hd5, sample_id, write_pngs, stats, continuous_stats) -> None:
    for xml_field in xml_field_ids:
        xmlp = xml_folder + str(sample_id) + '_' + xml_field + '*.xml'
        ecgs = glob.glob(xmlp)
        if xml_field == ECG_REST_FIELD:
            _write_ecg_rest_tensors(ecgs, xml_field, hd5, sample_id, write_pngs, stats, continuous_stats)
        elif xml_field == ECG_BIKE_FIELD:
            _write_ecg_bike_tensors(ecgs, xml_field, hd5, sample_id, stats)
        else:
            raise ValueError('Unknown ECG field ID:', xml_field)


def _write_ecg_rest_tensors(ecgs, xml_field, hd5, sample_id, write_pngs, stats, continuous_stats) -> None:
    rest_group = 'ukb_ecg_rest'
    for ecg in ecgs:
        logging.info('Got ECG for sample:{} XML field:{}'.format(sample_id, xml_field))
        root = et.parse(ecg).getroot()
        ecg_date = _str2date(_date_str_from_ecg(root))
        diagnosis_text = []
        for d in root.findall("./Interpretation/Diagnosis/DiagnosisText"):
            if 'QRS Complexes:' in d.text:
                qrs = float(d.text.replace('QRS Complexes:', '').split(',')[0].strip())
                create_tensor_in_hd5(hd5, rest_group, 'QRSComplexes', qrs, stats, date=ecg_date)
            elif '---' in d.text or 'Arrhythmia results of the full-disclosure ECG' in d.text:
                continue
            else:
                diagnosis_text.append(d.text.replace(',', '').replace('*', '').replace('&', 'and').replace('  ', ' '))

        diagnosis_str = ' '.join(diagnosis_text)
        create_tensor_in_hd5(hd5, rest_group, 'ecg_rest_text', diagnosis_str, stats, date=ecg_date, storage_type=StorageType.STRING)

        for c in root.findall("./StripData/WaveformData"):
            lead_data = list(map(float, c.text.strip().split(',')))
            dataset_name = 'strip_' + str(c.attrib['lead'])
            create_tensor_in_hd5(hd5, rest_group, dataset_name, lead_data, stats, date=ecg_date)
            stats[dataset_name] += 1

        for c in root.findall("./RestingECGMeasurements"):
            for child in c:
                if child.text is not None and child.tag in ECG_TAGS_TO_WRITE:
                    create_tensor_in_hd5(hd5, rest_group, child.tag, float(child.text), stats, date=ecg_date)
                    stats[child.tag] += 1
                    if write_pngs:
                        continuous_stats[child.tag].append(float(child.text))
                if child.tag == 'MedianSamples':
                    for median_c in child:
                        if median_c.tag == 'WaveformData':
                            median_wave = list(map(float, median_c.text.strip().split(',')))
                            dataset_name = 'median_' + str(median_c.attrib['lead'])
                            create_tensor_in_hd5(hd5, 'ukb_ecg_rest', dataset_name, median_wave, stats, date=ecg_date)
        for c in root.findall("./RestingECGMeasurements/MeasurementTable"):
            for child in c:
                if child.tag not in ECG_TABLE_TAGS:
                    continue
                values = list(map(_to_float_or_nan, child.text.strip().split(',')))
                create_tensor_in_hd5(hd5, 'ukb_ecg_rest', child.tag.lower(), values, stats, date=ecg_date)


def create_tensor_in_hd5(
    hd5: h5py.File, path_prefix: str, name: str, value, stats: Counter = None, date: datetime.datetime = None,
    storage_type: StorageType = None, attributes: Dict[str, Any] = None, instance: str = None,
):
    hd5_path = tensor_path(path_prefix, name)

    if instance is not None:
        hd5_path = f'{hd5_path}instance_{instance}'
    elif hd5_path in hd5:
        hd5_path = f'{hd5_path}instance_{len(hd5[hd5_path])}'
    else:
        hd5_path = f'{hd5_path}instance_0'

    if stats is not None:
        stats[hd5_path] += 1
    if storage_type == StorageType.STRING:
        d = hd5.create_dataset(hd5_path, data=value, dtype=h5py.special_dtype(vlen=str))
    elif isinstance(value, int) or isinstance(value, float):
        d = hd5.create_dataset(hd5_path, data=[value])
    elif isinstance(value, np.ndarray) or isinstance(value, list):
        d = hd5.create_dataset(hd5_path, data=value, compression='gzip')
    else:
        raise NotImplementedError(f'{storage_type} cannot be automatically written yet')  # TODO: Add categorical, etc.
    if date is not None:
        d.attrs['date'] = time.mktime(date.timetuple())
    if attributes is not None:
        for k in attributes:
            d.attrs[k] = attributes[k]


def tensor_path(path_prefix: str, name: str) -> str:
    """
    In the future, TMAPs should be generated using this same function
    """
    return f'/{path_prefix}/{name}/'


def first_dataset_at_path(hd5, path, gather_fxn=min):
    if path not in hd5:
        raise ValueError(f'Could not find key:{path} in hd5.')
    data = hd5[path]
    if isinstance(data, h5py.Dataset):
        return data
    deeper_key_prefix = f'{path}{gather_fxn(hd5[path])}/'
    return first_dataset_at_path(hd5, deeper_key_prefix)


def _datetime_to_str(dt: datetime.datetime) -> str:
    return dt.strftime(DATE_FORMAT)


def path_date_to_datetime(date: str) -> datetime.datetime:
    return datetime.datetime.strptime(date, DATE_FORMAT)


def _write_ecg_bike_tensors(ecgs, xml_field, hd5, sample_id, stats):
    for ecg in ecgs:
        root = et.parse(ecg).getroot()
        date = datetime.datetime.strptime(_date_str_from_ecg(root), '%Y-%m-%d')
        instance = ecg.split(JOIN_CHAR)[-2]
        write_to_hd5 = partial(create_tensor_in_hd5, hd5=hd5, path_prefix='ukb_ecg_bike', instance=instance, stats=stats, date=date)
        logging.info(f'Got ECG for sample:{sample_id} XML field:{xml_field}')

        protocol = root.findall('./Protocol/Phase')[0].find('ProtocolName').text
        write_to_hd5(storage_type=StorageType.STRING, name='protocol', value=protocol)

        median_ecgs = defaultdict(list)
        for median_waves in root.findall('./MedianData/Median/WaveformData'):
            median_ecgs[median_waves.attrib['lead']].extend(list(map(float, median_waves.text.strip().split(','))))
        if len(median_ecgs) > 0:
            median_np = np.zeros(ECG_BIKE_MEDIAN_SIZE)
            for lead in median_ecgs:
                median_idx = min(ECG_BIKE_MEDIAN_SIZE[0], len(median_ecgs[lead]))
                median_np[:median_idx, ECG_BIKE_LEADS[lead]] = median_ecgs[lead][:median_idx]
            write_to_hd5(name='median', value=median_np)
        else:
            stats['missing median bike ECG'] += 1

        counter = 0
        strip_np = np.zeros(ECG_BIKE_STRIP_SIZE)
        for strip_waves in root.findall("./StripData/Strip/WaveformData"):
            counter += 1
            strip_list = list(map(float, strip_waves.text.strip().split(',')))
            strip_idx = min(ECG_BIKE_MEDIAN_SIZE[0], len(strip_list))
            strip_np[:strip_idx, ECG_BIKE_LEADS[strip_waves.attrib['lead']]] = strip_list[:strip_idx]
        if counter > 0:
            write_to_hd5(name='strip', value=strip_np)
        else:
            stats['missing strip bike ECG'] += 1

        full_ekgs = [[] for _ in range(ECG_BIKE_NUM_LEADS)]
        count = 0
        for full_d in root.findall("./FullDisclosure/FullDisclosureData"):
            for full_line in re.split('\n|\t', full_d.text):
                for sample in re.split(',', full_line):
                    if sample == '':
                        continue
                    lead = (count % (ECG_BIKE_NUM_LEADS * ECG_BIKE_SAMPLE_RATE)) // ECG_BIKE_SAMPLE_RATE
                    full_ekgs[lead].append(float(sample))
                    count += 1

        if all(full_ekgs):  # Does each lead have data?
            full_np = np.zeros(ECG_BIKE_FULL_SIZE)
            for i, lead in enumerate(full_ekgs):
                full_idx = min(ECG_BIKE_FULL_SIZE[0], len(lead))
                full_np[:full_idx, i] = lead[:full_idx]
            write_to_hd5(name='full', value=full_np)
        else:
            stats['missing full disclosure bike ECG'] += 1

        # Patient info
        patient_fields = ('Age', 'Height', 'Weight')
        for field in patient_fields:
            val = [_xml_path_to_float(root, f'./PatientInfo/{field}')]
            write_to_hd5(name=str.lower(field), value=val)

        # Trend measurements
        trend_entry_fields = ['HeartRate', 'Load', 'Grade', 'Mets', 'VECount', 'PaceCount']
        trend_lead_measurements = [
            'JPointAmplitude',
            'STAmplitude20ms',
            'STAmplitude',
            'RAmplitude',
            'R1Amplitude',
            'STSlope',
            'STIntegral',
        ]
        phase_to_int = {'Pretest': 0, 'Exercise': 1, 'Rest': 2}
        lead_to_int = {'I': 0, '2': 1, '3': 2}
        trend_entries = root.findall("./TrendData/TrendEntry")

        trends = {trend: np.full(len(trend_entries), np.nan) for trend in trend_entry_fields + ['time', 'PhaseTime', 'PhaseTime', 'PhaseName', 'Artifact']}
        trends.update({trend: np.full((len(trend_entries), 3), np.nan) for trend in trend_lead_measurements})

        for i, trend_entry in enumerate(trend_entries):
            for field in trend_entry_fields:
                field_val = trend_entry.find(field)
                if field_val is None:
                    continue
                field_val = _to_float_or_false(field_val.text)
                if field_val is False:
                    continue
                trends[field][i] = field_val
            for lead_field, lead in product(trend_lead_measurements, trend_entry.findall('LeadMeasurements')):
                lead_num = lead.attrib['lead']
                field_val = lead.find(lead_field)
                if field_val is None:
                    continue
                field_val = _to_float_or_false(field_val.text.strip('?'))
                if field_val is False:
                    continue
                trends[lead_field][i, lead_to_int[lead_num]] = field_val
            try:
                trends['time'][i] = SECONDS_PER_MINUTE * int(trend_entry.find("EntryTime/Minute").text) + int(trend_entry.find("EntryTime/Second").text)
                trends['PhaseTime'][i] = SECONDS_PER_MINUTE * int(trend_entry.find("PhaseTime/Minute").text) + int(trend_entry.find("PhaseTime/Second").text)
                trends['PhaseName'][i] = phase_to_int[trend_entry.find('PhaseName').text]
                trends['Artifact'][i] = float(trend_entry.find('Artifact').text.strip('%')) / 100  # Artifact is reported as a percentage
            except AttributeError as e:
                stats['AttributeError on Trend Data'] += 1

        for field, trend_list in trends.items():
            write_to_hd5(name=f'trend_{str.lower(field)}', value=trend_list)

        # Last 60 seconds of raw given that the rest phase is 60s
        phase_durations = {}
        for protocol in root.findall("./Protocol/Phase"):
            phase_name = protocol.find("PhaseName").text
            phase_duration = SECONDS_PER_MINUTE * int(protocol.find("PhaseDuration/Minute").text) + int(
                protocol.find("PhaseDuration/Second").text,
            )
            phase_durations[phase_name] = phase_duration
            write_to_hd5(name=f'{str.lower(phase_name)}_duration', value=[phase_duration])

        # HR stats
        try:
            max_hr = _xml_path_to_float(root, './ExerciseMeasurements/MaxHeartRate')
            write_to_hd5(name='max_hr', value=[max_hr])
            resting_hr = _xml_path_to_float(root, './ExerciseMeasurements/RestingStats/RestHR')
            write_to_hd5(name='resting_hr', value=[resting_hr])
            max_pred_hr = _xml_path_to_float(root, './ExerciseMeasurements/MaxPredictedHR')
            write_to_hd5(name='max_pred_hr', value=[max_pred_hr])
        except AttributeError as e:
            stats['AttributeError on HR Stats'] += 1


def _write_tensors_from_niftis(folder: str, hd5: h5py.File, field_id: str, stats: Counter):
    niftis = glob.glob(os.path.join(folder, MRI_NIFTI_FIELD_ID_TO_ROOT[field_id], '**/*nii.gz'), recursive=True)
    logging.info(f'Found {len(niftis)} NIFTI files at {os.path.join(folder, MRI_NIFTI_FIELD_ID_TO_ROOT[field_id])} ')
    for nifti in niftis:  # iterate through all nii.gz files and add them to the hd5
        nifti_mri = nib.load(nifti)
        nifti_array = nifti_mri.get_fdata()
        nii_name = os.path.basename(nifti).replace('.nii.gz', '')  # removes .nii.gz
        parent = nifti.split('/')[-2]
        if parent not in nii_name:
            nii_name = f'{parent}_{nii_name}'
        stats[nii_name] += 1
        stats[f'{nii_name} shape:{nifti_array.shape}'] += 1
        if MRI_NIFTI_FIELD_ID_TO_ROOT[field_id] == 'SWI' and nifti_array.shape[-1] == MRI_SWI_SLICES_TO_AXIS_SHIFT and nifti_array.shape[0] > nifti_array.shape[1]:
            nifti_array = np.moveaxis(nifti_array, 0, 1)
            stats[f'{nii_name} shape post SWI shift:{nifti_array.shape}'] += 1
        for i in range(nifti_array.shape[-1]):
            create_tensor_in_hd5(hd5=hd5, path_prefix=f'ukb_brain_mri/{nii_name}', name=f'axial_{i}', value=nifti_array[..., i], stats=stats)


def _xml_path_to_float(root: et, path: str) -> float:
    return float(root.find(path).text)


def _date_str_from_ecg(root):
    date_str = ''
    for d in root.findall('./ObservationDateTime/Year'):
        date_str += d.text + '-'
    for d in root.findall('./ObservationDateTime/Month'):
        date_str += d.text + '-'
    for d in root.findall('./ObservationDateTime/Day'):
        date_str += d.text
    return date_str


def _get_disease_censor_dates(disease2tsv, min_sample_id=0, max_sample_id=9e9) -> Dict[str, Dict[int, List[datetime.datetime]]]:
    censor_dates = defaultdict(dict)
    for d in disease2tsv:
        with open(disease2tsv[d], 'r') as my_tsv:
            lol = list(csv.reader(my_tsv, delimiter='\t'))
            for row in lol[1:]:
                sample_id = int(row[0])
                if min_sample_id <= sample_id <= max_sample_id:
                    censor_dates[d][sample_id] = _str2date(row[4])
    return censor_dates


def disease_censor_status(disease2tsv, min_sample_id=0, max_sample_id=9e9) -> Dict[str, Dict[int, List[int]]]:
    censor_status = defaultdict(dict)
    for d in disease2tsv:
        with open(disease2tsv[d], 'r') as my_tsv:
            lol = list(csv.reader(my_tsv, delimiter='\t'))
            for row in lol[1:]:
                sample_id = int(row[0])
                if min_sample_id <= sample_id <= max_sample_id:
                    censor_status[d][sample_id] = 0 if row[1] == 'NA' else int(row[1])
    return censor_status


def disease_prevalence_status(disease2tsv, min_sample_id=0, max_sample_id=9e9) -> Dict[str, Dict[int, List[int]]]:
    censor_status = defaultdict(dict)
    for d in disease2tsv:
        with open(disease2tsv[d], 'r') as my_tsv:
            lol = list(csv.reader(my_tsv, delimiter='\t'))
            for row in lol[1:]:
                sample_id = int(row[0])
                if min_sample_id <= sample_id <= max_sample_id:
                    censor_status[d][sample_id] = 0 if row[2] == 'NA' else int(row[2])
    return censor_status


def disease_incidence_status(disease2tsv, min_sample_id=0, max_sample_id=9e9) -> Dict[str, Dict[int, List[int]]]:
    censor_status = defaultdict(dict)
    for d in disease2tsv:
        with open(disease2tsv[d], 'r') as my_tsv:
            lol = list(csv.reader(my_tsv, delimiter='\t'))
            for row in lol[1:]:
                sample_id = int(row[0])
                if min_sample_id <= sample_id <= max_sample_id:
                    censor_status[d][sample_id] = 0 if row[3] == 'NA' else int(row[3])
    return censor_status


def get_disease2tsv(tsv_folder) -> Dict[str, str]:
    ukb_prefix = 'ukb9222_'
    ukb_postfix = '_phenoV1.tsv'
    disease2tsv = {}
    tsvs = os.listdir(tsv_folder)
    for tsv in tsvs:
        disease_name = tsv.replace(ukb_prefix, '').replace(ukb_postfix, '').lower()
        disease2tsv[disease_name] = tsv_folder + tsv
    return disease2tsv


def append_fields_from_csv(tensors, csv_file, group, delimiter):
    stats = Counter()
    data_maps = defaultdict(dict)
    categorical_channel_maps = {MRI_ANNOTATION_NAME: MRI_ANNOTATION_CHANNEL_MAP}  # str: dict
    with open(csv_file, 'r') as volumes:
        lol = list(csv.reader(volumes, delimiter=delimiter))
        fields = lol[0][1:]  # Assumes sample id is the first field
        logging.info(f"CSV has {len(fields)} fields:{fields}")
        for row in lol[1:]:
            sample_id = row[0]
            data_maps[sample_id] = {fields[i]: row[i+1] for i in range(len(fields))}

    logging.info(f"Data maps:{len(data_maps)} for group:{group}")
    for tp in os.listdir(tensors):
        if os.path.splitext(tp)[-1].lower() != TENSOR_EXT:
            continue
        try:
            with h5py.File(tensors + tp, 'a') as hd5:
                sample_id = tp.replace(TENSOR_EXT, '')
                if sample_id in data_maps:
                    stats['Samples found'] += 1
                    if stats['Samples found'] % 250 == 0:
                        for k in stats:
                            logging.info(f'{k} has: {stats[k]}')
                    for field in data_maps[sample_id]:
                        value = data_maps[sample_id][field]
                        if group == 'continuous':
                            try:
                                value = float(value.strip())
                            except ValueError:
                                stats[f'could not cast field: {field} with value: {value} to float'] += 1
                                continue
                        elif group == 'categorical':
                            is_channel_mapped = False
                            for cm_name in categorical_channel_maps:
                                if value in categorical_channel_maps[cm_name]:
                                    hd5_key = group + HD5_GROUP_CHAR + cm_name
                                    if cm_name in hd5[group]:
                                        data = hd5[hd5_key]
                                        data[0] = categorical_channel_maps[cm_name][value]
                                        stats['updated'] += 1
                                    else:
                                        hd5.create_dataset(hd5_key, data=[categorical_channel_maps[cm_name][value]])
                                        stats['created'] += 1
                                    is_channel_mapped = True
                            if is_channel_mapped:
                                continue

                            if value.lower() in ['false', 'f', '0']:
                                value = 0
                            elif value.lower() in ['true', 't', '1']:
                                value = 1
                            else:
                                stats[f'Could not parse categorical field: {field} with value: {value}'] += 1
                                continue

                        hd5_key = group + HD5_GROUP_CHAR + field
                        if field in hd5[group]:
                            data = hd5[hd5_key]
                            data[0] = value
                            stats['updated'] += 1
                        else:
                            hd5.create_dataset(hd5_key, data=[value])
                            stats['created'] += 1
                else:
                    stats['sample id missing']
        except:
            print('could not open', tp, traceback.format_exc())
            stats['failed'] += 1

    logging.info(f'Finished appending data!')
    for k in stats:
        logging.info(f'{k} has: {stats[k]}')


def append_gene_csv(tensors, csv_file, delimiter):
    stats = Counter()
    data_maps = defaultdict(dict)
    with open(csv_file, 'r') as volumes:
        lol = list(csv.reader(volumes, delimiter=delimiter))
        fields = lol[0][1:]  # Assumes sample id is the first field
        logging.info(f"CSV of flag data header:{fields}")
        for row in lol[1:]:
            sample_id = row[0]
            data_maps[sample_id] = {fields[i]: row[i+1] for i in range(len(fields))}

    logging.info(f"Data maps:{len(data_maps)}")
    hd5_prefix = 'categorical' + HD5_GROUP_CHAR
    for tp in os.listdir(tensors):
        if os.path.splitext(tp)[-1].lower() != TENSOR_EXT:
            continue
        try:
            with h5py.File(tensors + tp, 'a') as hd5:
                sample_id = tp.replace(TENSOR_EXT, '')
                if sample_id in data_maps:
                    for field in data_maps[sample_id]:
                        if field in hd5[hd5_prefix]:
                            data = hd5[data_maps[sample_id][field]]
                            data[0] = 1.0
                            stats['updated'] += 1
                        else:
                            hd5.create_dataset(hd5_prefix + field, data=[1.0])
                            stats['created'] += 1
                else:
                    stats['sample id missing']
        except:
            print('couldnt open', tp, traceback.format_exc())
            stats['failed'] += 1

    for k in stats:
        logging.info("{}: {}".format(k, stats[k]))


# TODO Use 'with' or explicitly close files opened in this method
def _ukbb_stats(run_id, output_folder, phenos_folder, volume_csv, icd_csv, app_csv, zip_folder) -> None:
    stats = Counter()

    steve2sek = {}
    lol = list(csv.reader(open(app_csv, 'r'), delimiter=','))
    for row in lol[1:]:
        steve2sek[row[0]] = row[1]

    disease2tsv = get_disease2tsv(phenos_folder)
    logging.info('got disease tsvs:{}'.format(disease2tsv))
    dates = _get_disease_censor_dates(disease2tsv, 1000000, 2000000)
    status = disease_censor_status(disease2tsv, 1000000, 2000000)
    lol = list(csv.reader(open(volume_csv, 'r'), delimiter='\t'))
    logging.info(list(enumerate(lol[0])))
    lvesv = {}
    lvedv = {}
    lvef = {}
    for row in lol[1:]:
        sample_id = int(row[0])
        lvesv[sample_id] = float(row[2])
        lvedv[sample_id] = float(row[4])
        lvef[sample_id] = float(row[6])

    lol = list(csv.reader(open(icd_csv, 'r'), delimiter='\t'))
    logging.info('CSV of ICDs header:{}'.format(list(enumerate(lol[0]))))

    icd_indexes = {}
    icds = defaultdict(dict)
    for i, header in enumerate(lol[0]):
        if header.lower() in disease2tsv:
            icd_indexes[header.lower()] = i
    logging.info('got icd indexes:{}'.format(icd_indexes))

    disease_dates = defaultdict(list)
    for row in lol[1:]:
        sample_id = int(row[0])
        if os.path.exists(zip_folder + row[0] + '_20209_2_0.zip') \
                or os.path.exists(zip_folder + row[0] + '_20209_0_0.zip') \
                or os.path.exists(zip_folder + row[0] + '_20209_1_0.zip'):
            stats['sample_with_sax_mri'] += 1
        for disease in disease2tsv:
            if sample_id not in dates[disease]:
                stats['sample_id but no status or dates '] += 1
                continue
            if disease not in icd_indexes:
                continue
            icds[disease][sample_id] = row[icd_indexes[disease]]
            if icds[disease][sample_id] != 'NA' and sample_id in status[disease]:
                if status[disease][sample_id] != int(icds[disease][sample_id]):
                    stats[
                        disease + '_icd_disparity_' + icds[disease][sample_id] + '_censor_' + str(
                        status[disease][sample_id],
                        )
                    ] += 1
                else:
                    stats[disease + '_icd_agree'] += 1

            icds[disease][sample_id] = status[disease][sample_id]
            if icds[disease][sample_id] != 'NA' and int(icds[disease][sample_id]) == 1:
                stats[disease + '_total'] += 1
                if sample_id in dates[disease]:
                    stats[disease + '_with_date'] += 1
                    if dates[disease][sample_id] > _str2date('1940-11-11'):
                        disease_dates[disease].append(dates[disease][sample_id])
                if sample_id in lvef:
                    stats[disease + '_with_mri'] += 1
                if sample_id in lvef and sample_id in dates[disease]:
                    stats[disease + '_with_mri_and_date'] += 1

    for k in sorted(list(stats.keys())):
        logging.info('{} has: {}'.format(k, stats[k]))

    logging.info('Plot dates for diseases:{}'.format(list(disease_dates.keys())))
    for i, d in enumerate(disease_dates):
        figure_path = os.path.join(output_folder, run_id, d + '_dates' + IMAGE_EXT)
        if not os.path.exists(os.path.dirname(figure_path)):
            os.makedirs(os.path.dirname(figure_path))
        logging.info('Try to plot figure at:{} with {} dates.'.format(figure_path, len(disease_dates[d])))
        plt.figure(figsize=(16, 16))
        plt.title(d)
        plt.hist(disease_dates[d], bins=60)
        plt.savefig(figure_path)
        plt.close()

    # rows = max(1, int(math.ceil(math.sqrt(len(disease_dates)))))
    # cols = math.ceil(len(disease_dates)/rows)
    # fig, axes = plt.subplots(rows, cols, figsize=(28,28))
    # for i,d in enumerate(disease_dates):
    #     ax = plt.subplot(rows, cols, i+1)
    #     ax.set_title(d)
    #     ax.hist(disease_dates[d], bins=60)
    # plt.savefig(output_folder + 'disease_dates.png')

    _log_extreme_n(lvef, 3)
    _log_extreme_n(lvedv, 3)
    _log_extreme_n(lvesv, 3)


def _print_disease_tensor_maps(phenos_folder) -> None:
    disease2tsv = get_disease2tsv(phenos_folder)
    status = disease_censor_status(disease2tsv)
    for d in sorted(list(disease2tsv.keys())):
        total = len(status[d])
        diseased = np.sum(list(status[d].values()))
        factor = int(total / (diseased * 2))
        print(
            "'{}': TensorMap('{}', group='categorical_index', channel_map={{'no_{}':0, '{}':1}}, loss=weighted_crossentropy([1.0, {}], '{}')),".format(
            d, d, d, d, factor, d,
            ),
        )


def _print_disease_tensor_maps_incident_prevalent(phenos_folder) -> None:
    disease2tsv = get_disease2tsv(phenos_folder)
    status_p = disease_prevalence_status(disease2tsv, 1000000, 2000000)
    status_i = disease_incidence_status(disease2tsv, 1000000, 2000000)
    for d in sorted(list(disease2tsv.keys())):
        total = len(status_p[d])
        diseased_p = np.sum(list(status_p[d].values()))
        factor_p = int(total / (1 + (diseased_p * 3)))
        diseased_i = np.sum(list(status_i[d].values()))
        factor_i = int(total / (1 + (diseased_i * 3)))
        print(
            "'{}_prevalent_incident': TensorMap('{}', group='categorical_date', channel_map={{'no_{}':0, 'prevalent_{}':1, 'incident_{}':2}}, loss=weighted_crossentropy([1.0, {}, {}], '{}_prevalent_incident')),".format(
                d, d, d, d, d, factor_p, factor_i, d,
            ),
        )


def _print_disease_tensor_maps_time(phenos_folder) -> None:
    disease2tsv = get_disease2tsv(phenos_folder)
    disease_tm_str = "'{}_time': TensorMap('{}', group='diagnosis_time', channel_map={{'{}_time':0}}, loss='mse'),"
    for d in sorted(list(disease2tsv.keys())):
        print(disease_tm_str.format(d, d, d))


def _plot_mi_hospital_only(db, run_id, output_folder) -> None:
    conn = sqlite3.connect(db)
    sql_cursor = conn.cursor()
    q = "SELECT datething.value FROM phenotype datething where datething.fieldid=42000 and sample_id in  (select sample_id from phenotype where FieldID=42001 and value=0);"
    dates = []
    for data_row in sql_cursor.execute(q):
        dates.append(_str2date(data_row[0]))
    plt.figure(figsize=(12, 12))
    plt.xlabel('MI Date')
    plt.hist(dates, bins=60)
    plt.savefig(os.path.join(output_folder, run_id, 'mi_self_report_dates' + IMAGE_EXT))


def _str2date(d) -> datetime.date:
    parts = d.split(CONCAT_CHAR)
    if len(parts) < 2:
        return datetime.datetime.now().date()
    return datetime.date(int(parts[0]), int(parts[1]), int(parts[2]))


def _date_from_dicom(d) -> str:
    return d.AcquisitionDate[0:4] + CONCAT_CHAR + d.AcquisitionDate[4:6] + CONCAT_CHAR + d.AcquisitionDate[6:]


def _datetime_from_dicom(d) -> datetime.date:
    return _str2date(_date_from_dicom(d))


def _log_extreme_n(stats, n) -> None:
    logging.info('Min values:')
    i = 0
    ordered = sorted(stats.items(), key=operator.itemgetter(1))
    for k, v in ordered:
        logging.info('{} has: {}'.format(k, v))
        i += 1
        if i > n:
            break
    logging.info('Max values:')
    i = 0
    for k, v in ordered[-n:-1]:
        logging.info('{} has: {}'.format(k, v))
        i += 1
        if i > n:
            break
    logging.info('\n\n')


def _prune_sample(
    sample_id: int, min_sample_id: int, max_sample_id: int, mri_field_ids: List[int],
    xml_field_ids: List[int], zip_folder: str, xml_folder: str,
):
    """Return True if the sample ID is missing associated MRI, EKG, or GT data.  Or if the sample_id is below the given minimum."""

    if sample_id < min_sample_id:
        return True
    if sample_id > max_sample_id:
        return True
    if len(mri_field_ids) > 0 and not (_sample_has_dicom_mris(zip_folder, sample_id) or _sample_has_nifti_mris(zip_folder, sample_id)):
        return True
    if len(xml_field_ids) > 0 and not _sample_has_ecgs(xml_folder, xml_field_ids, sample_id):
        return True

    return False
